#!/usr/bin/env  python
__license__   = 'GPL v3'
__copyright__ = '2008, Kovid Goyal kovid@kovidgoyal.net'
__docformat__ = 'restructuredtext en'

'''
Command line interface to the calibre database.
'''

import sys, os, cStringIO, re
import unicodedata
from textwrap import TextWrapper

from calibre import preferred_encoding, prints, isbytestring, patheq
from calibre.constants import iswindows
from calibre.db.legacy import LibraryDatabase
from calibre.utils.config import OptionParser, prefs, tweaks
from calibre.ebooks.metadata.meta import get_metadata
from calibre.ebooks.metadata.book.base import field_from_string
from calibre.ebooks.metadata.opf2 import OPFCreator, OPF
from calibre.utils.date import isoformat
from calibre.utils.localization import canonicalize_lang

FIELDS = set(['title', 'authors', 'author_sort', 'publisher', 'rating',
    'timestamp', 'size', 'tags', 'comments', 'series', 'series_index',
    'formats', 'isbn', 'uuid', 'pubdate', 'cover', 'last_modified',
    'identifiers'])

do_notify = True
def send_message(msg=''):
    global do_notify
    if not do_notify:
        return
    prints('Notifying calibre of the change')
    from calibre.utils.ipc import RC
    t = RC(print_error=False)
    t.start()
    t.join(3)
    if t.done:
        t.conn.send('refreshdb:'+msg)
        t.conn.close()

def write_dirtied(db):
    prints('Backing up metadata')
    db.dump_metadata()

def get_parser(usage):
    parser = OptionParser(usage)
    go = parser.add_option_group(_('GLOBAL OPTIONS'))
    go.add_option('--library-path', '--with-library', default=None, help=_('Path to the calibre library. Default is to use the path stored in the settings.'))
    go.add_option('--dont-notify-gui', default=False, action='store_true',
            help=_('Do not notify the running calibre GUI (if any) that the database has'
                ' changed. Use with care, as it can lead to database corruption!'))

    return parser

def get_db(dbpath, options):
    global do_notify
    if options.library_path is not None:
        dbpath = os.path.expanduser(options.library_path)
    if dbpath is None:
        raise ValueError('No saved library path, either run the GUI or use the'
                ' --with-library option')
    dbpath = os.path.abspath(dbpath)
    if options.dont_notify_gui:
        do_notify = False
    return LibraryDatabase(dbpath)

def do_list(db, fields, afields, sort_by, ascending, search_text, line_width, separator,
            prefix, limit, for_machine=False):
    from calibre.utils.terminal import ColoredStream, geometry
    if sort_by:
        db.sort(sort_by, ascending)
    if search_text:
        db.search(search_text)
    data = db.get_data_as_dict(prefix, authors_as_string=True, convert_to_local_tz=False)
    if limit > -1:
        data = data[:limit]
    fields = ['id'] + fields
    title_fields = fields
    def field_name(f):
        ans = f
        if f[0] == '*':
            if f.endswith('_index'):
                fkey = f[1:-len('_index')]
                num = db.custom_column_label_map[fkey]['num']
                ans = '%d_index'%num
            else:
                ans = db.custom_column_label_map[f[1:]]['num']
        return ans
    if for_machine:
        import json
        for record in data:
            for key in set(record) - set(fields):
                del record[key]
            for key, val in tuple(record.iteritems()):
                if hasattr(val, 'isoformat'):
                    record[key] = isoformat(val, as_utc=True)
                elif val is None:
                    del record[key]
        return json.dumps(data, indent=2, sort_keys=True)
    fields = list(map(field_name, fields))

    for f in data:
        fmts = [x for x in f['formats'] if x is not None]
        f['formats'] = u'[%s]'%u', '.join(fmts)
    widths = list(map(lambda x: 0, fields))
    for record in data:
        for f in record.keys():
            if hasattr(record[f], 'isoformat'):
                record[f] = isoformat(record[f], as_utc=False)
            else:
                record[f] = unicode(record[f])
            record[f] = record[f].replace('\n', ' ')
    def chr_width(x):
        return 1 + unicodedata.east_asian_width(x).startswith('W')
    def str_width(x):
        return sum(map(chr_width, x))

    for i in data:
        for j, field in enumerate(fields):
            widths[j] = max(widths[j], str_width(i[field]))

    screen_width = geometry()[0] if line_width < 0 else line_width
    if not screen_width:
        screen_width = 80
    field_width = screen_width//len(fields)
    base_widths = map(lambda x: min(x+1, field_width), widths)

    while sum(base_widths) < screen_width:
        adjusted = False
        for i in range(len(widths)):
            if base_widths[i] < widths[i]:
                base_widths[i] += min(screen_width-sum(base_widths), widths[i]-base_widths[i])
                adjusted = True
                break
        if not adjusted:
            break

    widths = list(base_widths)
    titles = map(lambda x, y: '%-*s%s'%(x-len(separator), y, separator),
            widths, title_fields)
    with ColoredStream(sys.stdout, fg='green'):
        print ''.join(titles)

    wrappers = [TextWrapper(x - 1).wrap if x > 1 else lambda y: y for x in widths]
    o = cStringIO.StringIO()

    for record in data:
        text = [wrappers[i](unicode(record[field])) for i, field in enumerate(fields)]
        lines = max(map(len, text))
        for l in range(lines):
            for i, field in enumerate(text):
                ft = text[i][l] if l < len(text[i]) else u''
                o.write(ft.encode('utf-8'))
                if i < len(text) - 1:
                    filler = (u'%*s'%(widths[i]-str_width(ft)-1, u''))
                    o.write((filler+separator).encode('utf-8'))
            print >>o
    return o.getvalue()

def list_option_parser(db=None):
    fields = set(FIELDS)
    if db is not None:
        for f, data in db.custom_column_label_map.iteritems():
            fields.add('*'+f)
            if data['datatype'] == 'series':
                fields.add('*'+f+'_index')

    parser = get_parser(_(
'''\
%prog list [options]

List the books available in the calibre database.
'''
                            ))
    parser.add_option('-f', '--fields', default='title,authors',
                      help=_('The fields to display when listing books in the'
                          ' database. Should be a comma separated list of'
                          ' fields.\nAvailable fields: %s\nDefault: %%default. The'
                          ' special field "all" can be used to select all fields.'
                          )%', '.join(sorted(fields)))
    parser.add_option('--sort-by', default=None,
                      help=_('The field by which to sort the results.\nAvailable fields: %s\nDefault: %%default')%','.join(sorted(FIELDS)))
    parser.add_option('--ascending', default=False, action='store_true',
                      help=_('Sort results in ascending order'))
    parser.add_option('-s', '--search', default=None,
                      help=_('Filter the results by the search query. For the format of the search query,'
                             ' please see the search related documentation in the User Manual. Default is to do no filtering.'))
    parser.add_option('-w', '--line-width', default=-1, type=int,
                      help=_('The maximum width of a single line in the output. Defaults to detecting screen size.'))
    parser.add_option('--separator', default=' ', help=_('The string used to separate fields. Default is a space.'))
    parser.add_option('--prefix', default=None, help=_('The prefix for all file paths. Default is the absolute path to the library folder.'))
    parser.add_option('--limit', default=-1, type=int, help=_('The maximum number of results to display. Default: all'))
    parser.add_option('--for-machine', default=False, action='store_true', help=_(
        'Generate output in JSON format, which is more suitable for machine parsing. Causes the line width and separator options to be ignored.'))
    return parser


def command_list(args, dbpath):
    pre = get_parser('')
    pargs = [x for x in args if x.startswith('--with-library') or x.startswith('--library-path')
        or not x.startswith('-')]
    opts = pre.parse_args(sys.argv[:1] + pargs)[0]
    db = get_db(dbpath, opts)
    parser = list_option_parser(db=db)
    opts, args = parser.parse_args(sys.argv[:1] + args)
    afields = set(FIELDS)
    if db is not None:
        for f, data in db.custom_column_label_map.iteritems():
            afields.add('*'+f)
            if data['datatype'] == 'series':
                afields.add('*'+f+'_index')
    if opts.fields.strip():
        fields = [str(f.strip().lower()) for f in opts.fields.split(',')]
        if 'all' in fields:
            fields = sorted(list(afields))
        if not set(fields).issubset(afields):
            parser.print_help()
            print
            prints(_('Invalid fields. Available fields:'),
                    ','.join(sorted(afields)), file=sys.stderr)
            return 1
    else:
        fields = []

    if opts.sort_by not in afields and opts.sort_by is not None:
        parser.print_help()
        print
        prints(_('Invalid sort field. Available fields:'), ','.join(afields),
                file=sys.stderr)
        return 1

    print do_list(db, fields, afields, opts.sort_by, opts.ascending, opts.search, opts.line_width, opts.separator,
            opts.prefix, opts.limit, for_machine=opts.for_machine)
    return 0


class DevNull(object):

    def write(self, msg):
        pass
NULL = DevNull()

def do_add(db, paths, one_book_per_directory, recurse, add_duplicates, otitle,
        oauthors, oisbn, otags, oseries, oseries_index, ocover, olanguages):
    orig = sys.stdout
    # sys.stdout = NULL
    try:
        files, dirs = [], []
        for path in paths:
            path = os.path.abspath(path)
            if os.path.isdir(path):
                dirs.append(path)
            else:
                if os.path.exists(path):
                    files.append(path)
                else:
                    print path, 'not found'

        formats, metadata = [], []
        for book in files:
            format = os.path.splitext(book)[1]
            format = format[1:] if format else None
            if not format:
                continue
            stream = open(book, 'rb')
            mi = get_metadata(stream, stream_type=format, use_libprs_metadata=True)
            if not mi.title:
                mi.title = os.path.splitext(os.path.basename(book))[0]
            if not mi.authors:
                mi.authors = [_('Unknown')]
            for x in ('title', 'authors', 'isbn', 'tags', 'series', 'languages'):
                val = locals()['o'+x]
                if val:
                    setattr(mi, x, val)
            if oseries:
                mi.series_index = oseries_index
            if ocover:
                mi.cover = ocover

            formats.append(format)
            metadata.append(mi)

        file_duplicates = []
        added_ids = set()
        if files:
            file_duplicates, ids = db.add_books(files, formats, metadata,
                                           add_duplicates=add_duplicates,
                                           return_ids=True)
            added_ids |= set(ids)

        dir_dups = []
        for dir in dirs:
            if recurse:
                dir_dups.extend(db.recursive_import(dir,
                    single_book_per_directory=one_book_per_directory,
                    added_ids=added_ids))
            else:
                func = db.import_book_directory if one_book_per_directory else db.import_book_directory_multiple
                dups = func(dir, added_ids=added_ids)
                if not dups:
                    dups = []
                dir_dups.extend(dups)

        sys.stdout = sys.__stdout__

        if add_duplicates:
            for mi, formats in dir_dups:
                book_id = db.import_book(mi, formats)
                added_ids.add(book_id)
        else:
            if dir_dups or file_duplicates:
                print >>sys.stderr, _('The following books were not added as '
                                      'they already exist in the database '
                                      '(see --duplicates option):')
            for mi, formats in dir_dups:
                title = mi.title
                if isinstance(title, unicode):
                    title = title.encode(preferred_encoding)
                print >>sys.stderr, '\t', title + ':'
                for path in formats:
                    print >>sys.stderr, '\t\t ', path
            if file_duplicates:
                for path, mi in zip(file_duplicates[0], file_duplicates[2]):
                    title = mi.title
                    if isinstance(title, unicode):
                        title = title.encode(preferred_encoding)
                    print >>sys.stderr, '\t', title+':'
                    print >>sys.stderr, '\t\t ', path

        write_dirtied(db)
        if added_ids:
            prints(_('Added book ids: %s')%(', '.join(map(type(u''),
                added_ids))))
        send_message()
    finally:
        sys.stdout = orig

def add_option_parser():
    parser = get_parser(_(
'''\
%prog add [options] file1 file2 file3 ...

Add the specified files as books to the database. You can also specify directories, see
the directory related options below.
'''
                            ))
    parser.add_option('-1', '--one-book-per-directory', action='store_true', default=False,
                      help=_('Assume that each directory has only a single logical book and that all files in it are different e-book formats of that book'))
    parser.add_option('-r', '--recurse', action='store_true', default=False,
                      help=_('Process directories recursively'))
    parser.add_option('-d', '--duplicates', action='store_true', default=False,
                      help=_('Add books to database even if they already exist. Comparison is done based on book titles.'))
    parser.add_option('-e', '--empty', action='store_true', default=False,
                    help=_('Add an empty book (a book with no formats)'))
    parser.add_option('-t', '--title', default=None,
            help=_('Set the title of the added book(s)'))
    parser.add_option('-a', '--authors', default=None,
            help=_('Set the authors of the added book(s)'))
    parser.add_option('-i', '--isbn', default=None,
            help=_('Set the ISBN of the added book(s)'))
    parser.add_option('-T', '--tags', default=None,
            help=_('Set the tags of the added book(s)'))
    parser.add_option('-s', '--series', default=None,
            help=_('Set the series of the added book(s)'))
    parser.add_option('-S', '--series-index', default=1.0, type=float,
            help=_('Set the series number of the added book(s)'))
    parser.add_option('-c', '--cover', default=None,
            help=_('Path to the cover to use for the added book'))
    parser.add_option('-l', '--languages', default=None,
            help=_('A comma separated list of languages (best to use ISO639 language codes, though some language names may also be recognized)'))

    return parser

def do_add_empty(db, title, authors, isbn, tags, series, series_index, cover, languages):
    from calibre.ebooks.metadata import MetaInformation
    mi = MetaInformation(None)
    if title is not None:
        mi.title = title
    if authors:
        mi.authors = authors
    if isbn:
        mi.isbn = isbn
    if tags:
        mi.tags = tags
    if series:
        mi.series, mi.series_index = series, series_index
    if cover:
        mi.cover = cover
    if languages:
        mi.languages = languages
    book_id = db.import_book(mi, [])
    write_dirtied(db)
    prints(_('Added book ids: %s')%book_id)
    send_message()

def command_add(args, dbpath):
    from calibre.ebooks.metadata import string_to_authors
    parser = add_option_parser()
    opts, args = parser.parse_args(sys.argv[:1] + args)
    aut = string_to_authors(opts.authors) if opts.authors else []
    tags = [x.strip() for x in opts.tags.split(',')] if opts.tags else []
    lcodes = [canonicalize_lang(x) for x in (opts.languages or '').split(',')]
    lcodes = [x for x in lcodes if x]
    if opts.empty:
        do_add_empty(get_db(dbpath, opts), opts.title, aut, opts.isbn, tags,
                opts.series, opts.series_index, opts.cover, lcodes)
        return 0
    if len(args) < 2:
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify at least one file to add')
        return 1
    do_add(get_db(dbpath, opts), args[1:], opts.one_book_per_directory,
            opts.recurse, opts.duplicates, opts.title, aut, opts.isbn,
            tags, opts.series, opts.series_index, opts.cover, lcodes)
    return 0

def do_remove(db, ids):
    book_ids = set()
    for x in ids:
        if isinstance(x, int):
            book_ids.add(x)
        else:
            book_ids |= set(x)

    db.new_api.remove_books(book_ids)
    db.clean()
    send_message()
    from calibre.db.delete_service import delete_service
    delete_service().wait()

def remove_option_parser():
    return get_parser(_(
'''\
%prog remove ids

Remove the books identified by ids from the database. ids should be a comma separated \
list of id numbers (you can get id numbers by using the list command). For example, \
23,34,57-85 (when specifying a range, the last number in the range is not
included).
'''))

def command_remove(args, dbpath):
    parser = remove_option_parser()
    opts, args = parser.parse_args(sys.argv[:1] + args)
    if len(args) < 2:
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify at least one book to remove')
        return 1

    ids = []
    for x in args[1].split(','):
        y = x.split('-')
        if len(y) > 1:
            ids.extend(range(int(y[0]), int(y[1])))
        else:
            ids.append(int(y[0]))

    do_remove(get_db(dbpath, opts), set(ids))

    return 0

def do_add_format(db, id, fmt, path, opts):
    done = db.add_format_with_hooks(id, fmt.upper(), path, index_is_id=True,
                             replace=opts.replace)
    if not done and not opts.replace:
        prints(_('A %(fmt)s file already exists for book: %(id)d, not replacing')%dict(fmt=fmt.upper(), id=id))
    else:
        send_message()

def add_format_option_parser():
    parser = get_parser(_(
'''\
%prog add_format [options] id ebook_file

Add the ebook in ebook_file to the available formats for the logical book identified \
by id. You can get id by using the list command. If the format already exists, \
it is replaced, unless the do not replace option is specified.\
'''))
    parser.add_option('--dont-replace', dest='replace', default=True, action='store_false',
                      help=_('Do not replace the format if it already exists'))
    return parser


def command_add_format(args, dbpath):
    parser = add_format_option_parser()
    opts, args = parser.parse_args(sys.argv[:1] + args)
    if len(args) < 3:
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify an id and an ebook file')
        return 1

    id, path, fmt = int(args[1]), args[2], os.path.splitext(args[2])[-1]
    if not fmt:
        print _('ebook file must have an extension')
    do_add_format(get_db(dbpath, opts), id, fmt[1:], path, opts)
    return 0

def do_remove_format(db, id, fmt):
    db.remove_format(id, fmt, index_is_id=True)
    send_message()
    from calibre.db.delete_service import delete_service
    delete_service().wait()

def remove_format_option_parser():
    return get_parser(_(
'''
%prog remove_format [options] id fmt

Remove the format fmt from the logical book identified by id. \
You can get id by using the list command. fmt should be a file extension \
like LRF or TXT or EPUB. If the logical book does not have fmt available, \
do nothing.
'''))


def command_remove_format(args, dbpath):
    parser = remove_format_option_parser()
    opts, args = parser.parse_args(sys.argv[:1] + args)
    if len(args) < 3:
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify an id and a format')
        return 1

    id, fmt = int(args[1]), args[2].upper()
    do_remove_format(get_db(dbpath, opts), id, fmt)
    return 0

def do_show_metadata(db, id, as_opf):
    if not db.has_id(id):
        raise ValueError('Id #%d is not present in database.'%id)
    mi = db.get_metadata(id, index_is_id=True)
    if as_opf:
        mi = OPFCreator(os.getcwdu(), mi)
        mi.render(sys.stdout)
    else:
        prints(unicode(mi))

def show_metadata_option_parser():
    parser = get_parser(_(
'''
%prog show_metadata [options] id

Show the metadata stored in the calibre database for the book identified by id.
id is an id number from the list command.
'''))
    parser.add_option('--as-opf', default=False, action='store_true',
                      help=_('Print metadata in OPF form (XML)'))
    return parser

def command_show_metadata(args, dbpath):
    parser = show_metadata_option_parser()
    opts, args = parser.parse_args(sys.argv[1:]+args)
    if len(args) < 2:
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify an id')
        return 1
    id = int(args[1])
    do_show_metadata(get_db(dbpath, opts), id, opts.as_opf)
    return 0

def do_set_metadata(db, id, stream):
    mi = OPF(stream).to_book_metadata()
    db.set_metadata(id, mi)

def set_metadata_option_parser():
    parser = get_parser(_(
'''
%prog set_metadata [options] id [/path/to/metadata.opf]

Set the metadata stored in the calibre database for the book identified by id
from the OPF file metadata.opf. id is an id number from the list command. You
can get a quick feel for the OPF format by using the --as-opf switch to the
show_metadata command. You can also set the metadata of individual fields with
the --field option. If you use the --field option, there is no need to specify
an OPF file.
'''))
    parser.add_option('-f', '--field', action='append', default=[], help=_(
        'The field to set. Format is field_name:value, for example: '
        '{0} tags:tag1,tag2. Use {1} to get a list of all field names. You '
        'can specify this option multiple times to set multiple fields. '
        'Note: For languages you must use the ISO639 language codes (e.g. '
        'en for English, fr for French and so on). For identifiers, the '
        'syntax is {0} {2}. For boolean (yes/no) fields use true and false '
        'or yes and no.'
        ).format('--field', '--list-fields', 'identifiers:isbn:XXXX,doi:YYYYY')
        )
    parser.add_option('-l', '--list-fields', action='store_true',
        default=False, help=_('List the metadata field names that can be used'
        ' with the --field option'))
    return parser

def embed_metadata_option_parser():
    parser = get_parser(_(
'''
%prog embed_metadata [options] book_id

Update the metadata in the actual book files stored in the calibre library from
the metadata in the calibre database.  Normally, metadata is updated only when
exporting files from calibre, this command is useful if you want the files to
be updated in place. Note that different file formats support different amounts
of metadata. You can use the special value 'all' for book_id to update metadata
in all books. You can also specify many book ids separated by spaces and id ranges
separated by hyphens. For example: %prog embed_metadata 1 2 10-15 23'''))
    parser.add_option('-f', '--only-formats', action='append', default=[], help=_(
        'Only update metadata in files of the specified format. Specify it multiple'
        ' times for multiple formats. By default, all formats are updated.'))
    return parser

def command_embed_metadata(args, dbpath):
    parser = embed_metadata_option_parser()
    opts, args = parser.parse_args(sys.argv[0:1]+args)
    db = get_db(dbpath, opts)
    ids = set()
    for x in args[1:]:
        if x == 'all':
            ids = db.new_api.all_book_ids()
            break
        parts = x.split('-')
        if len(parts) == 1:
            ids.add(int(parts[0]))
        else:
            ids |= {x for x in xrange(int(parts[0], int(parts[1])))}
    only_fmts = opts.only_formats or None
    def progress(i, total, mi):
        prints(_('Processed {0} ({1} of {2})').format(mi.title, i, total))
    db.new_api.embed_metadata(ids, only_fmts=only_fmts, report_progress=progress)
    send_message()

def command_set_metadata(args, dbpath):
    parser = set_metadata_option_parser()
    opts, args = parser.parse_args(sys.argv[0:1]+args)
    db = get_db(dbpath, opts)

    def fields():
        for key in sorted(db.field_metadata.all_field_keys()):
            m = db.field_metadata[key]
            if (key not in {'formats', 'series_sort', 'ondevice', 'path',
                'last_modified'} and m['is_editable'] and m['name']):
                yield key, m
                if m['datatype'] == 'series':
                    si = m.copy()
                    si['name'] = m['name'] + ' Index'
                    si['datatype'] = 'float'
                    yield key+'_index', si
        c = db.field_metadata['cover'].copy()
        c['datatype'] = 'text'
        yield 'cover', c

    if opts.list_fields:
        prints('%-40s'%_('Title'), _('Field name'), '\n')
        for key, m in fields():
            prints('%-40s'%m['name'], key)

        return 0

    def verify_int(x):
        try:
            int(x)
            return True
        except:
            return False

    if len(args) < 2 or not verify_int(args[1]):
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify a record id as the '
                'first argument')
        return 1
    if len(args) < 3 and not opts.field:
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify either a field or an opf file')
        return 1
    book_id = int(args[1])
    if book_id not in db.all_ids():
        prints(_('No book with id: %s in the database')%book_id, file=sys.stderr)
        raise SystemExit(1)

    if len(args) > 2:
        opf = args[2]
        if not os.path.exists(opf):
            prints(_('The OPF file %s does not exist')%opf, file=sys.stderr)
            return 1
        do_set_metadata(db, book_id, opf)

    if opts.field:
        fields = {k:v for k, v in fields()}
        fields['title_sort'] = fields['sort']
        vals = {}
        for x in opts.field:
            field, val = x.partition(':')[::2]
            if field == 'sort':
                field = 'title_sort'
            if field not in fields:
                print >>sys.stderr, _('%s is not a known field'%field)
                return 1
            val = field_from_string(field, val, fields[field])
            vals[field] = val
        mi = db.get_metadata(book_id, index_is_id=True, get_cover=False)
        for field, val in sorted(  # ensure series_index fields are set last
                vals.iteritems(), key=lambda k: 1 if k[0].endswith('_index') else 0):
            if field.endswith('_index'):
                try:
                    val = float(val)
                except Exception:
                    print >>sys.stderr, 'The value %r is not a valid series index' % val
                    raise SystemExit(1)
                sname = mi.get(field[:-6])
                if not sname:
                    print >>sys.stderr, 'Cannot set index for series before setting the series name'
                    raise SystemExit(1)
                mi.set(field[:-6], sname, extra=val)
                if field == 'series_index':
                    mi.series_index = val  # extra has no effect for the builtin series field
            else:
                mi.set(field, val)
        db.set_metadata(book_id, mi, force_changes=True)
    db.clean()
    do_show_metadata(db, book_id, False)
    write_dirtied(db)
    send_message()

    return 0

def do_export(db, ids, dir, opts):
    if ids is None:
        ids = list(db.all_ids())
    from calibre.library.save_to_disk import save_to_disk
    failures = save_to_disk(db, ids, dir, opts=opts)

    if failures:
        prints('Failed to save the following books:')
        for id, title, tb in failures:
            prints(str(id)+':', title)
            prints('\t'+'\n\t'.join(tb.splitlines()))
            prints(' ')

def export_option_parser():
    parser = get_parser(_('''\
%prog export [options] ids

Export the books specified by ids (a comma separated list) to the filesystem.
The export operation saves all formats of the book, its cover and metadata (in
an opf file). You can get id numbers from the list command.
'''))
    parser.add_option('--all', default=False, action='store_true',
                      help=_('Export all books in database, ignoring the list of ids.'))
    parser.add_option('--to-dir', default='.',
                      help=(_('Export books to the specified directory. Default is')+' %default'))
    parser.add_option('--single-dir', default=False, action='store_true',
                      help=_('Export all books into a single directory'))
    from calibre.library.save_to_disk import config
    c = config()
    for pref in ['asciiize', 'update_metadata', 'write_opf', 'save_cover']:
        opt = c.get_option(pref)
        switch = '--dont-'+pref.replace('_', '-')
        parser.add_option(switch, default=True, action='store_false',
                help=opt.help+' '+_('Specifying this switch will turn '
                    'this behavior off.'), dest=pref)

    for pref in ['timefmt', 'template', 'formats']:
        opt = c.get_option(pref)
        switch = '--'+pref
        parser.add_option(switch, default=opt.default,
                help=opt.help, dest=pref)

    for pref in ('replace_whitespace', 'to_lowercase'):
        opt = c.get_option(pref)
        switch = '--'+pref.replace('_', '-')
        parser.add_option(switch, default=False, action='store_true',
                help=opt.help)

    return parser

def command_export(args, dbpath):
    parser = export_option_parser()
    opts, args = parser.parse_args(sys.argv[1:]+args)
    if (len(args) < 2 and not opts.all):
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify some ids or the %s option')%'--all'
        return 1
    ids = None if opts.all else map(int, args[1].split(','))
    dir = os.path.abspath(os.path.expanduser(opts.to_dir))
    do_export(get_db(dbpath, opts), ids, dir, opts)
    return 0

def do_add_custom_column(db, label, name, datatype, is_multiple, display):
    num = db.create_custom_column(label, name, datatype, is_multiple, display=display)
    prints('Custom column created with id: %s'%num)

def add_custom_column_option_parser():
    from calibre.library.custom_columns import CustomColumns
    parser = get_parser(_('''\
%prog add_custom_column [options] label name datatype

Create a custom column. label is the machine friendly name of the column. Should
not contain spaces or colons. name is the human friendly name of the column.
datatype is one of: {0}
''').format(', '.join(sorted(CustomColumns.CUSTOM_DATA_TYPES))))

    parser.add_option('--is-multiple', default=False, action='store_true',
                      help=_('This column stores tag like data (i.e. '
                          'multiple comma separated values). Only '
                          'applies if datatype is text.'))
    parser.add_option('--display', default='{}',
            help=_('A dictionary of options to customize how '
                'the data in this column will be interpreted. This is a JSON '
                ' string. For enumeration columns, use '
                '--display="{\\"enum_values\\":[\\"val1\\", \\"val2\\"]}"'
                '\n'
                'There are many options that can go into the display variable.'
                'The options by column type are:\n'
                'composite: composite_template, composite_sort, make_category,'
                'contains_html, use_decorations\n'
                'datetime: date_format\n'
                'enumeration: enum_values, enum_colors, use_decorations\n'
                'int, float: number_format\n'
                'text: is_names, use_decorations\n'
                '\n'
                'The best way to find legal combinations is to create a custom '
                'column of the appropriate type in the GUI then look at the '
                'backup OPF for a book (ensure that a new OPF has been created '
                'since the column was added). You will see the JSON for the '
                '"display" for the new column in the OPF.'))
    return parser


def command_add_custom_column(args, dbpath):
    import json
    parser = add_custom_column_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) < 3:
        parser.print_help()
        print
        print >>sys.stderr, _('You must specify label, name and datatype')
        return 1
    do_add_custom_column(get_db(dbpath, opts), args[0], args[1], args[2],
            opts.is_multiple, json.loads(opts.display))
    # Re-open the DB so that  field_metadata is reflects the column changes
    db = get_db(dbpath, opts)
    db.prefs['field_metadata'] = db.field_metadata.all_metadata()
    return 0

def catalog_option_parser(args):
    from calibre.customize.ui import available_catalog_formats, plugin_for_catalog_format
    from calibre.utils.logging import Log

    def add_plugin_parser_options(fmt, parser, log):

        # Fetch the extension-specific CLI options from the plugin
        # library.catalogs.<format>.py
        plugin = plugin_for_catalog_format(fmt)
        for option in plugin.cli_options:
            if option.action:
                parser.add_option(option.option,
                                  default=option.default,
                                  dest=option.dest,
                                  action=option.action,
                                  help=option.help)
            else:
                parser.add_option(option.option,
                                  default=option.default,
                                  dest=option.dest,
                                  help=option.help)

        return plugin

    def validate_command_line(parser, args, log):
        # calibredb catalog path/to/destination.[epub|csv|xml|...] [options]

        # Validate form
        if not len(args) or args[0].startswith('-'):
            parser.print_help()
            log.error("\n\nYou must specify a catalog output file of the form 'path/to/destination.extension'\n"
            "To review options for an output format, type 'calibredb catalog <.extension> --help'\n"
            "For example, 'calibredb catalog .xml --help'\n")
            raise SystemExit(1)

        # Validate plugin exists for specified output format
        output = os.path.abspath(args[0])
        file_extension = output[output.rfind('.') + 1:].lower()

        if file_extension not in available_catalog_formats():
            parser.print_help()
            log.error("No catalog plugin available for extension '%s'.\n" % file_extension +
                      "Catalog plugins available for %s\n" % ', '.join(available_catalog_formats()))
            raise SystemExit(1)

        return output, file_extension

    # Entry point
    log = Log()
    parser = get_parser(_(
    '''
    %prog catalog /path/to/destination.(CSV|EPUB|MOBI|XML ...) [options]

    Export a catalog in format specified by path/to/destination extension.
    Options control how entries are displayed in the generated catalog output.
    '''))

    # Confirm that a plugin handler exists for specified output file extension
    # Will raise SystemExit(1) if no plugin matching file_extension
    output, fmt = validate_command_line(parser, args, log)

    # Add options common to all catalog plugins
    parser.add_option('-i', '--ids', default=None, dest='ids',
                      help=_("Comma-separated list of database IDs to catalog.\n"
                      "If declared, --search is ignored.\n"
                             "Default: all"))
    parser.add_option('-s', '--search', default=None, dest='search_text',
                      help=_("Filter the results by the search query. "
                          "For the format of the search query, please see "
                          "the search-related documentation in the User Manual.\n"
                      "Default: no filtering"))
    parser.add_option('-v','--verbose', default=False, action='store_true',
                      dest='verbose',
                      help=_('Show detailed output information. Useful for debugging'))

    # Add options specific to fmt plugin
    plugin = add_plugin_parser_options(fmt, parser, log)

    return parser, plugin, log

def command_catalog(args, dbpath):
    parser, plugin, log = catalog_option_parser(args)
    opts, args = parser.parse_args(sys.argv[1:])

    if len(args) < 2:
        parser.print_help()
        print
        print >>sys.stderr, _('Error: You must specify a catalog output file')
        return 1
    if opts.ids:
        opts.ids = [int(id) for id in opts.ids.split(',')]

    # No support for connected device in CLI environment
    # Parallel initialization in calibre.gui2.tools:generate_catalog()
    opts.connected_device = {
                             'is_device_connected': False,
                             'kind': None,
                             'name': None,
                             'save_template': None,
                             'serial': None,
                             'storage': None,
                            }

    with plugin:
        return int(bool(plugin.run(args[1], opts, get_db(dbpath, opts))))

def parse_series_string(db, label, value):
    val = unicode(value).strip()
    s_index = None
    pat = re.compile(r'\[([.0-9]+)\]')
    match = pat.search(val)
    if match is not None:
        val = pat.sub('', val).strip()
        s_index = float(match.group(1))
    elif val:
        if tweaks['series_index_auto_increment'] == 'no_change':
            pass
        elif tweaks['series_index_auto_increment'] != 'const':
            s_index = db.get_next_cc_series_num_for(val, label=label)
        else:
            s_index = 1.0
    return val, s_index

def do_set_custom(db, col, id_, val, append):
    if id_ not in db.all_ids():
        prints(_('No book with id: %s in the database')%id_, file=sys.stderr)
        raise SystemExit(1)
    if db.custom_column_label_map[col]['datatype'] == 'series':
        val, s_index = parse_series_string(db, col, val)
        db.set_custom(id_, val, extra=s_index, label=col, append=append)
        prints('Data set to: %r[%4.2f]'%
               (db.get_custom(id_, label=col, index_is_id=True),
                db.get_custom_extra(id_, label=col, index_is_id=True)))
    else:
        db.set_custom(id_, val, label=col, append=append)
        prints('Data set to: %r'%db.get_custom(id_, label=col, index_is_id=True))

def set_custom_option_parser():
    parser = get_parser(_(
    '''
    %prog set_custom [options] column id value

    Set the value of a custom column for the book identified by id.
    You can get a list of ids using the list command.
    You can get a list of custom column names using the custom_columns
    command.
    '''))

    parser.add_option('-a', '--append', default=False, action='store_true',
            help=_('If the column stores multiple values, append the specified '
                'values to the existing ones, instead of replacing them.'))
    return parser


def command_set_custom(args, dbpath):
    parser = set_custom_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) < 3:
        parser.print_help()
        print
        print >>sys.stderr, _('Error: You must specify a field name, id and value')
        return 1
    do_set_custom(get_db(dbpath, opts), args[0], int(args[1]), args[2],
            opts.append)
    return 0

def do_custom_columns(db, details):
    from pprint import pformat
    cols = db.custom_column_label_map
    for col, data in cols.items():
        if details:
            prints(col)
            print
            prints(pformat(data))
            print '\n'
        else:
            prints(col, '(%d)'%data['num'])

def custom_columns_option_parser():
    parser = get_parser(_(
    '''
    %prog custom_columns [options]

    List available custom columns. Shows column labels and ids.
    '''))
    parser.add_option('-d', '--details', default=False, action='store_true',
            help=_('Show details for each column.'))
    return parser


def command_custom_columns(args, dbpath):
    parser = custom_columns_option_parser()
    opts, args = parser.parse_args(args)
    do_custom_columns(get_db(dbpath, opts), opts.details)
    return 0

def do_remove_custom_column(db, label, force):
    if not force:
        q = raw_input(_('You will lose all data in the column: %s.'
            ' Are you sure (y/n)? ')%label)
        if q.lower().strip() != _('y'):
            return
    try:
        db.delete_custom_column(label=label)
    except KeyError:
        prints(_('No column named %s found. You must use column labels, not titles.'
               ' Use calibredb custom_columns to get a list of labels.')%label, file=sys.stderr)
        raise SystemExit(1)
    prints('Column %r removed.'%label)

def remove_custom_column_option_parser():
    parser = get_parser(_(
    '''
    %prog remove_custom_column [options] label

    Remove the custom column identified by label. You can see available
    columns with the custom_columns command.
    '''))
    parser.add_option('-f', '--force', default=False, action='store_true',
            help=_('Do not ask for confirmation'))
    return parser


def command_remove_custom_column(args, dbpath):
    parser = remove_custom_column_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) < 1:
        parser.print_help()
        print
        prints(_('Error: You must specify a column label'), file=sys.stderr)
        return 1

    do_remove_custom_column(get_db(dbpath, opts), args[0], opts.force)
    # Re-open the DB so that  field_metadata is reflects the column changes
    db = get_db(dbpath, opts)
    db.prefs['field_metadata'] = db.field_metadata.all_metadata()
    return 0

def saved_searches_option_parser():
    parser = get_parser(_(
    '''
    %prog saved_searches [options] list
    %prog saved_searches add name search
    %prog saved_searches remove name

    Manage the saved searches stored in this database.
    If you try to add a query with a name that already exists, it will be
    replaced.
    '''))
    return parser

def command_saved_searches(args, dbpath):
    parser = saved_searches_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) < 1:
        parser.print_help()
        print
        prints(_('Error: You must specify an action (add|remove|list)'), file=sys.stderr)
        return 1
    db = get_db(dbpath, opts)
    if args[0] == 'list':
        for name in db.saved_search_names():
            prints(_('Name:'), name)
            prints(_('Search string:'), db.saved_search_lookup(name))
            print
    elif args[0] == 'add':
        if len(args) < 3:
            parser.print_help()
            print
            prints(_('Error: You must specify a name and a search string'), file=sys.stderr)
            return 1
        db.saved_search_add(args[1], args[2])
        prints(args[1], _('added'))
    elif args[0] == 'remove':
        if len(args) < 2:
            parser.print_help()
            print
            prints(_('Error: You must specify a name'), file=sys.stderr)
            return 1
        db.saved_search_delete(args[1])
        prints(args[1], _('removed'))
    else:
        parser.print_help()
        print
        prints(_('Error: Action %s not recognized, must be one '
            'of: (add|remove|list)') % args[1], file=sys.stderr)
        return 1

    return 0

def backup_metadata_option_parser():
    parser = get_parser(_('''\
%prog backup_metadata [options]

Backup the metadata stored in the database into individual OPF files in each
books directory. This normally happens automatically, but you can run this
command to force re-generation of the OPF files, with the --all option.

Note that there is normally no need to do this, as the OPF files are backed up
automatically, every time metadata is changed.
'''))
    parser.add_option('--all', default=False, action='store_true',
        help=_('Normally, this command only operates on books that have'
            ' out of date OPF files. This option makes it operate on all'
            ' books.'))
    return parser

class BackupProgress(object):

    def __init__(self):
        self.total = 0
        self.count = 0

    def __call__(self, book_id, mi, ok):
        if mi is True:
            self.total = book_id
        else:
            self.count += 1
            prints(u'%.1f%% %s - %s'%((self.count*100)/float(self.total),
                book_id, mi.title))

def command_backup_metadata(args, dbpath):
    parser = backup_metadata_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) != 0:
        parser.print_help()
        return 1

    if opts.library_path is not None:
        dbpath = opts.library_path
    if isbytestring(dbpath):
        dbpath = dbpath.decode(preferred_encoding)
    db = LibraryDatabase(dbpath)
    book_ids = None
    if opts.all:
        book_ids = db.all_ids()
    db.dump_metadata(book_ids=book_ids, callback=BackupProgress())


def check_library_option_parser():
    from calibre.library.check_library import CHECKS
    parser = get_parser(_('''\
%prog check_library [options]

Perform some checks on the filesystem representing a library. Reports are {0}
''').format(', '.join([c[0] for c in CHECKS])))

    parser.add_option('-c', '--csv', default=False, action='store_true',
            help=_('Output in CSV'))

    parser.add_option('-r', '--report', default=None, dest='report',
                      help=_("Comma-separated list of reports.\n"
                             "Default: all"))

    parser.add_option('-e', '--ignore_extensions', default=None, dest='exts',
                      help=_("Comma-separated list of extensions to ignore.\n"
                             "Default: all"))

    parser.add_option('-n', '--ignore_names', default=None, dest='names',
                      help=_("Comma-separated list of names to ignore.\n"
                             "Default: all"))
    return parser

def command_check_library(args, dbpath):
    from calibre.library.check_library import CheckLibrary, CHECKS
    parser = check_library_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) != 0:
        parser.print_help()
        return 1

    if opts.library_path is not None:
        dbpath = opts.library_path

    if isbytestring(dbpath):
        dbpath = dbpath.decode(preferred_encoding)

    if opts.report is None:
        checks = CHECKS
    else:
        checks = []
        for r in opts.report.split(','):
            found = False
            for c in CHECKS:
                if c[0] == r:
                    checks.append(c)
                    found = True
                    break
            if not found:
                print _('Unknown report check'), r
                return 1

    if opts.names is None:
        names = []
    else:
        names = [f.strip() for f in opts.names.split(',') if f.strip()]
    if opts.exts is None:
        exts = []
    else:
        exts = [f.strip() for f in opts.exts.split(',') if f.strip()]

    def print_one(checker, check):
        attr = check[0]
        list = getattr(checker, attr, None)
        if list is None:
            return
        if opts.csv:
            for i in list:
                print check[1] + ',' + i[0] + ',' + i[1]
        else:
            print check[1]
            for i in list:
                print '    %-40.40s - %-40.40s'%(i[0], i[1])

    if not LibraryDatabase.exists_at(dbpath):
        prints('No library found at', dbpath, file=sys.stderr)
        raise SystemExit(1)

    db = LibraryDatabase(dbpath)
    checker = CheckLibrary(dbpath, db)
    checker.scan_library(names, exts)
    for check in checks:
        print_one(checker, check)


def restore_database_option_parser():
    parser = get_parser(_(
    '''\
%prog restore_database [options]

Restore this database from the metadata stored in OPF files in each
directory of the calibre library. This is useful if your metadata.db file
has been corrupted.

WARNING: This command completely regenerates your database. You will lose
all saved searches, user categories, plugboards, stored per-book conversion
settings, and custom recipes. Restored metadata will only be as accurate as
what is found in the OPF files.
    '''))

    parser.add_option('-r', '--really-do-it', default=False, action='store_true',
            help=_('Really do the recovery. The command will not run '
                   'unless this option is specified.'))
    return parser

def command_restore_database(args, dbpath):
    parser = restore_database_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) != 0:
        parser.print_help()
        return 1

    if not opts.really_do_it:
        prints(_('You must provide the %s option to do a'
            ' recovery')%'--really-do-it', end='\n\n')
        parser.print_help()
        return 1

    if opts.library_path is not None:
        dbpath = opts.library_path

    if isbytestring(dbpath):
        dbpath = dbpath.decode(preferred_encoding)

    class Progress(object):
        def __init__(self):
            self.total = 1

        def __call__(self, msg, step):
            if msg is None:
                self.total = float(step)
            else:
                prints(msg, '...', '%d%%'%int(100*(step/self.total)))
    from calibre.db.restore import Restore
    r = Restore(dbpath, progress_callback=Progress())
    r.start()
    r.join()

    if r.tb is not None:
        prints('Restoring database failed with error:')
        prints(r.tb)
    else:
        prints('Restoring database succeeded')
        prints('old database saved as', r.olddb)
        if r.errors_occurred:
            name = 'calibre_db_restore_report.txt'
            open('calibre_db_restore_report.txt',
                    'wb').write(r.report.encode('utf-8'))
            prints('Some errors occurred. A detailed report was '
                    'saved to', name)

def list_categories_option_parser():
    parser = get_parser(_('''\
%prog list_categories [options]

Produce a report of the category information in the database. The
information is the equivalent of what is shown in the tags pane.
'''))

    parser.add_option('-i', '--item_count', default=False, action='store_true',
            help=_('Output only the number of items in a category instead of the '
                   'counts per item within the category'))
    parser.add_option('-c', '--csv', default=False, action='store_true',
            help=_('Output in CSV'))
    parser.add_option('-q', '--quote', default='"',
            help=_('The character to put around the category value in CSV mode. '
                   'Default is quotes (").'))
    parser.add_option('-r', '--categories', default='', dest='report',
                      help=_("Comma-separated list of category lookup names.\n"
                             "Default: all"))
    parser.add_option('-w', '--width', default=-1, type=int,
                      help=_('The maximum width of a single line in the output. '
                             'Defaults to detecting screen size.'))
    parser.add_option('-s', '--separator', default=',',
                      help=_('The string used to separate fields in CSV mode. '
                             'Default is a comma.'))
    return parser

def command_list_categories(args, dbpath):
    parser = list_categories_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) != 0:
        parser.print_help()
        return 1

    if opts.library_path is not None:
        dbpath = opts.library_path

    if isbytestring(dbpath):
        dbpath = dbpath.decode(preferred_encoding)

    db = LibraryDatabase(dbpath)
    category_data = db.get_categories()
    data = []
    report_on = [c.strip() for c in opts.report.split(',') if c.strip()]
    categories = [k for k in category_data.keys()
                  if db.metadata_for_field(k)['kind'] not in ['user', 'search'] and
                  (not report_on or k in report_on)]

    categories.sort(cmp=lambda x,y: cmp(x if x[0] != '#' else x[1:],
                                        y if y[0] != '#' else y[1:]))
    if not opts.item_count:
        for category in categories:
            is_rating = db.metadata_for_field(category)['datatype'] == 'rating'
            for tag in category_data[category]:
                if is_rating:
                    tag.name = unicode(len(tag.name))
                data.append({'category':category, 'tag_name':tag.name,
                             'count':unicode(tag.count), 'rating':unicode(tag.avg_rating)})
    else:
        for category in categories:
            data.append({'category':category,
                         'tag_name':_('CATEGORY ITEMS'),
                         'count': len(category_data[category]), 'rating': 0.0})

    fields = ['category', 'tag_name', 'count', 'rating']

    def do_list():
        from calibre.utils.terminal import geometry, ColoredStream

        separator = ' '
        widths = list(map(lambda x: 0, fields))
        for i in data:
            for j, field in enumerate(fields):
                widths[j] = max(widths[j], max(len(field), len(unicode(i[field]))))

        screen_width = geometry()[0]
        if not screen_width:
            screen_width = 80
        field_width = screen_width//len(fields)
        base_widths = map(lambda x: min(x+1, field_width), widths)

        while sum(base_widths) < screen_width:
            adjusted = False
            for i in range(len(widths)):
                if base_widths[i] < widths[i]:
                    base_widths[i] += min(screen_width-sum(base_widths), widths[i]-base_widths[i])
                    adjusted = True
                    break
            if not adjusted:
                break

        widths = list(base_widths)
        titles = map(lambda x, y: '%-*s%s'%(x-len(separator), y, separator),
                widths, fields)
        with ColoredStream(sys.stdout, fg='green'):
            print ''.join(titles)

        wrappers = map(lambda x: TextWrapper(x-1), widths)
        o = cStringIO.StringIO()

        for record in data:
            text = [wrappers[i].wrap(unicode(record[field]).encode('utf-8')) for i, field in enumerate(fields)]
            lines = max(map(len, text))
            for l in range(lines):
                for i, field in enumerate(text):
                    ft = text[i][l] if l < len(text[i]) else ''
                    filler = '%*s'%(widths[i]-len(ft)-1, '')
                    o.write(ft)
                    o.write(filler+separator)
                print >>o
        print o.getvalue()

    def do_csv():
        lf = '{category},"{tag_name}",{count},{rating}'
        lf = lf.replace(',', opts.separator).replace(r'\t','\t').replace(r'\n','\n')
        lf = lf.replace('"', opts.quote)
        for d in data:
            print lf.format(**d)

    if opts.csv:
        do_csv()
    else:
        do_list()

    return parser

def clone_option_parser():
    return get_parser(_(
    '''\
%prog clone path/to/new/library

Create a clone of the current library. This creates a new, empty library that has all the
same custom columns, virtual libraries and other settings as the current library.

The cloned library will contain no books. If you want to create a full duplicate, including
all books, then simply use your filesystem tools to copy the library folder.
    '''))

def command_clone(args, dbpath):
    parser = clone_option_parser()
    opts, args = parser.parse_args(args)
    if len(args) < 1:
        parser.print_help()
        print
        prints(_('Error: You must specify the path to the cloned library'))
        return 1
    db = get_db(dbpath, opts)
    loc = args[0]
    if not os.path.exists(loc):
        os.makedirs(loc)
    loc = os.path.abspath(loc)

    if patheq(loc, db.library_path):
        prints(_('The location for the new library is the same as the current library'))
        return 1
    empty = not os.listdir(loc)
    if not empty:
        prints(_('%s is not empty. You must choose an empty directory for the new library.') % loc)
        return 1
    if iswindows and len(loc) > LibraryDatabase.WINDOWS_LIBRARY_PATH_LIMIT:
        prints(_('Path to library too long. Must be less than'
                    ' %d characters.')%LibraryDatabase.WINDOWS_LIBRARY_PATH_LIMIT)
        return 1
    dbprefs = dict(db.prefs)
    db.close()
    LibraryDatabase(loc, default_prefs=dbprefs)


COMMANDS = ('list', 'add', 'remove', 'add_format', 'remove_format',
            'show_metadata', 'set_metadata', 'export', 'catalog',
            'saved_searches', 'add_custom_column', 'custom_columns',
            'remove_custom_column', 'set_custom', 'restore_database',
            'check_library', 'list_categories', 'backup_metadata',
            'clone', 'embed_metadata')


def option_parser():
    parser = OptionParser(_(
'''\
%%prog command [options] [arguments]

%%prog is the command line interface to the calibre books database.

command is one of:
  %s

For help on an individual command: %%prog command --help
'''
                          )%'\n  '.join(COMMANDS))
    return parser


def main(args=sys.argv):
    parser = option_parser()
    if len(args) < 2:
        parser.print_help()
        return 1
    if args[1] not in COMMANDS:
        if args[1] == '--version':
            parser.print_version()
            return 0
        parser.print_help()
        return 1

    command = eval('command_'+args[1])
    dbpath = prefs['library_path']

    return command(args[2:], dbpath)

if __name__ == '__main__':
    sys.exit(main())




__license__   = 'GPL v3'
__copyright__ = '2008, Kovid Goyal <kovid at kovidgoyal.net>'

from PyQt5.Qt import QDialog, QListWidgetItem, QModelIndex

from calibre.gui2 import file_icon_provider
from calibre.gui2.dialogs.choose_format_ui import Ui_ChooseFormatDialog

class ChooseFormatDialog(QDialog, Ui_ChooseFormatDialog):

    def __init__(self, window, msg, formats):
        QDialog.__init__(self, window)
        Ui_ChooseFormatDialog.__init__(self)
        self.setupUi(self)
        self.formats.activated[QModelIndex].connect(self.activated_slot)

        self.msg.setText(msg)
        for format in formats:
            self.formats.addItem(QListWidgetItem(file_icon_provider().icon_from_ext(format.lower()),
                                                 format.upper()))
        self._formats = formats
        self.formats.setCurrentRow(0)
        self._format = None

    def book_converted(self, book_id, fmt):
        fmt = fmt.upper()
        if fmt not in self._formats:
            self._formats.append(fmt)
            self.formats.addItem(QListWidgetItem(
                file_icon_provider().icon_from_ext(fmt.lower()), fmt.upper()))

    def activated_slot(self, *args):
        self.accept()

    def format(self):
        return self._format

    def accept(self):
        self._format = self._formats[self.formats.currentRow()]
        return QDialog.accept(self)




#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai

__license__   = 'GPL v3'
__copyright__ = '2010, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import os

from calibre.constants import plugins, filesystem_encoding

_magick, _merr = plugins['magick']

if _magick is None:
    raise RuntimeError('Failed to load ImageMagick: '+_merr)

_gravity_map = dict([(getattr(_magick, x), x) for x in dir(_magick) if
    x.endswith('Gravity')])

_type_map = dict([(getattr(_magick, x), x) for x in dir(_magick) if
    x.endswith('Type')])

_colorspace_map = dict([(getattr(_magick, x), x) for x in dir(_magick) if
    x.endswith('Colorspace')])

# Font metrics {{{
class Rect(object):

    def __init__(self, left, top, right, bottom):
        self.left, self.top, self.right, self.bottom = left, top, right, bottom

    def __str__(self):
        return '(%s, %s) -- (%s, %s)'%(self.left, self.top, self.right,
                self.bottom)

class FontMetrics(object):

    def __init__(self, ret):
        self._attrs = []
        for i, x in enumerate(('char_width', 'char_height', 'ascender',
            'descender', 'text_width', 'text_height',
            'max_horizontal_advance')):
            setattr(self, x, ret[i])
            self._attrs.append(x)
        self.bounding_box = Rect(ret[7], ret[8], ret[9], ret[10])
        self.x, self.y = ret[11], ret[12]
        self._attrs.extend(['bounding_box', 'x', 'y'])
        self._attrs = tuple(self._attrs)

    def __str__(self):
        return '''FontMetrics:
            char_width: %s
            char_height: %s
            ascender: %s
            descender: %s
            text_width: %s
            text_height: %s
            max_horizontal_advance: %s
            bounding_box: %s
            x: %s
            y: %s
            '''%tuple([getattr(self, x) for x in self._attrs])

# }}}

class PixelWand(_magick.PixelWand):  # {{{
    pass

# }}}

class DrawingWand(_magick.DrawingWand):  # {{{

    @dynamic_property
    def font(self):
        def fget(self):
            return self.font_.decode(filesystem_encoding, 'replace').lower()
        def fset(self, val):
            if isinstance(val, unicode):
                val = val.encode(filesystem_encoding)
            self.font_ = str(val)
        return property(fget=fget, fset=fset, doc=_magick.DrawingWand.font_.__doc__)

    @dynamic_property
    def gravity(self):
        def fget(self):
            val = self.gravity_
            return _gravity_map[val]
        def fset(self, val):
            val = getattr(_magick, str(val))
            self.gravity_ = val
        return property(fget=fget, fset=fset, doc=_magick.DrawingWand.gravity_.__doc__)

    @dynamic_property
    def font_size(self):
        def fget(self):
            return self.font_size_
        def fset(self, val):
            self.font_size_ = float(val)
        return property(fget=fget, fset=fset, doc=_magick.DrawingWand.font_size_.__doc__)

    @dynamic_property
    def stroke_color(self):
        def fget(self):
            return self.stroke_color_.color
        def fset(self, val):
            col = PixelWand()
            col.color = unicode(val)
            self.stroke_color_ = col
        return property(fget=fget, fset=fset, doc=_magick.DrawingWand.font_size_.__doc__)

    @dynamic_property
    def fill_color(self):
        def fget(self):
            return self.fill_color_.color
        def fset(self, val):
            col = PixelWand()
            col.color = unicode(val)
            self.fill_color_ = col
        return property(fget=fget, fset=fset, doc=_magick.DrawingWand.font_size_.__doc__)

# }}}

class Image(_magick.Image):  # {{{

    @property
    def clone(self):
        ans = Image()
        ans.copy(self)
        return ans

    def load(self, data):
        if not data:
            raise ValueError('Cannot open image from empty data string')
        data = bytes(data)
        return _magick.Image.load(self, data)

    def open(self, path_or_file):
        if not hasattr(path_or_file, 'read') and \
            path_or_file.lower().endswith('.wmf'):
            # Special handling for WMF files as ImageMagick seems
            # to hang while reading them from a blob on linux
            if isinstance(path_or_file, unicode):
                path_or_file = path_or_file.encode(filesystem_encoding)
            return _magick.Image.read(self, path_or_file)
        data = path_or_file
        if hasattr(data, 'read'):
            data = data.read()
        else:
            data = open(data, 'rb').read()
        if not data:
            raise ValueError('%r is an empty file'%path_or_file)
        self.load(data)

    @dynamic_property
    def format(self):
        def fget(self):
            return self.format_.decode('utf-8', 'ignore').lower()
        def fset(self, val):
            self.format_ = str(val)
        return property(fget=fget, fset=fset, doc=_magick.Image.format_.__doc__)

    @dynamic_property
    def type(self):
        def fget(self):
            return _type_map[self.type_]
        def fset(self, val):
            val = getattr(_magick, str(val))
            self.type_ = val
        return property(fget=fget, fset=fset, doc=_magick.Image.type_.__doc__)

    @dynamic_property
    def colorspace(self):
        def fget(self):
            return _colorspace_map[self.colorspace_]
        def fset(self, val):
            val = getattr(_magick, str(val))
            self.colorspace_ = val
        return property(fget=fget, fset=fset, doc=_magick.Image.type_.__doc__)

    @dynamic_property
    def size(self):
        def fget(self):
            return self.size_
        def fset(self, val):
            filter = 'CatromFilter'
            if len(val) > 2:
                filter = val[2]
            filter = int(getattr(_magick, filter))
            blur = 1.0
            if len(val) > 3:
                blur = float(val[3])
            self.size_ = (int(val[0]), int(val[1]), filter, blur)
        return property(fget=fget, fset=fset, doc=_magick.Image.size_.__doc__)

    def save(self, path, format=None):
        if format is None:
            ext = os.path.splitext(path)[1]
            if len(ext) < 2:
                raise ValueError('No format specified')
            format = ext[1:]
        format = format.upper()

        with lopen(path, 'wb') as f:
            f.write(self.export(format))

    def compose(self, img, left=0, top=0, operation='OverCompositeOp'):
        op = getattr(_magick, operation)
        bounds = self.size
        if left < 0 or top < 0 or left >= bounds[0] or top >= bounds[1]:
            raise ValueError('left and/or top out of bounds')
        _magick.Image.compose(self, img, int(left), int(top), op)

    def compare(self, img, metric='RootMeanSquaredErrorMetric'):
        return _magick.Image.compare(self, img, getattr(_magick, metric))

    def font_metrics(self, drawing_wand, text):
        if isinstance(text, unicode):
            text = text.encode('UTF-8')
        return FontMetrics(_magick.Image.font_metrics(self, drawing_wand, text))

    def annotate(self, drawing_wand, x, y, angle, text):
        if isinstance(text, unicode):
            text = text.encode('UTF-8')
        return _magick.Image.annotate(self, drawing_wand, x, y, angle, text)

    def distort(self, method, arguments, bestfit):
        method = getattr(_magick, method)
        arguments = [float(x) for x in arguments]
        _magick.Image.distort(self, method, arguments, bestfit)

    def rotate(self, background_pixel_wand, degrees):
        _magick.Image.rotate(self, background_pixel_wand, float(degrees))

    def quantize(self, number_colors, colorspace='RGBColorspace', treedepth=0, dither=True,
            measure_error=False):
        colorspace = getattr(_magick, colorspace)
        _magick.Image.quantize(self, number_colors, colorspace, treedepth, dither,
                measure_error)

    def trim(self, fuzz):
        try:
            _magick.Image.remove_border(self, fuzz)
        except AttributeError:
            _magick.Image.trim(self, fuzz)


# }}}

def create_canvas(width, height, bgcolor='#ffffff'):
    canvas = Image()
    canvas.create_canvas(int(width), int(height), str(bgcolor))
    return canvas



#!/usr/bin/env python
# vim:fileencoding=utf-8
from __future__ import (unicode_literals, division, absolute_import,
                        print_function)

__license__ = 'GPL v3'
__copyright__ = '2013, Kovid Goyal <kovid at kovidgoyal.net>'

from collections import OrderedDict
from calibre.ebooks.docx.names import XPath, get

class Inherit:
    pass
inherit = Inherit()

def binary_property(parent, name):
    vals = XPath('./w:%s' % name)(parent)
    if not vals:
        return inherit
    val = get(vals[0], 'w:val', 'on')
    return True if val in {'on', '1', 'true'} else False

def simple_color(col, auto='black'):
    if not col or col == 'auto' or len(col) != 6:
        return auto
    return '#'+col

def simple_float(val, mult=1.0):
    try:
        return float(val) * mult
    except (ValueError, TypeError, AttributeError, KeyError):
        return None


LINE_STYLES = {  # {{{
    'basicBlackDashes': 'dashed',
    'basicBlackDots': 'dotted',
    'basicBlackSquares': 'dashed',
    'basicThinLines': 'solid',
    'dashDotStroked': 'groove',
    'dashed': 'dashed',
    'dashSmallGap': 'dashed',
    'dotDash': 'dashed',
    'dotDotDash': 'dashed',
    'dotted': 'dotted',
    'double': 'double',
    'inset': 'inset',
    'nil': 'none',
    'none': 'none',
    'outset': 'outset',
    'single': 'solid',
    'thick': 'solid',
    'thickThinLargeGap': 'double',
    'thickThinMediumGap': 'double',
    'thickThinSmallGap' : 'double',
    'thinThickLargeGap': 'double',
    'thinThickMediumGap': 'double',
    'thinThickSmallGap': 'double',
    'thinThickThinLargeGap': 'double',
    'thinThickThinMediumGap': 'double',
    'thinThickThinSmallGap': 'double',
    'threeDEmboss': 'ridge',
    'threeDEngrave': 'groove',
    'triple': 'double',
}  # }}}

# Read from XML {{{

border_props = ('padding_%s', 'border_%s_width', 'border_%s_style', 'border_%s_color')

def read_single_border(parent, edge):
    color = style = width = padding = None
    for elem in XPath('./w:%s' % edge)(parent):
        c = get(elem, 'w:color')
        if c is not None:
            color = simple_color(c)
        s = get(elem, 'w:val')
        if s is not None:
            style = LINE_STYLES.get(s, 'solid')
        space = get(elem, 'w:space')
        if space is not None:
            try:
                padding = float(space)
            except (ValueError, TypeError):
                pass
        sz = get(elem, 'w:sz')
        if sz is not None:
            # we dont care about art borders (they are only used for page borders)
            try:
                # WebKit needs at least 1pt to render borders
                width = min(96, max(8, float(sz))) / 8
            except (ValueError, TypeError):
                pass
    if style == 'double' and width is not None and 0 < width < 3:
        width = 3  # WebKit needs 3pts to render double borders
    return {p:v for p, v in zip(border_props, (padding, width, style, color))}

def read_border(parent, dest, border_edges=('left', 'top', 'right', 'bottom'), name='pBdr'):
    vals = {k % edge:inherit for edge in border_edges for k in border_props}

    for border in XPath('./w:' + name)(parent):
        for edge in border_edges:
            for prop, val in read_single_border(border, edge).iteritems():
                if val is not None:
                    vals[prop % edge] = val

    for key, val in vals.iteritems():
        setattr(dest, key, val)

def read_indent(parent, dest):
    padding_left = padding_right = text_indent = inherit
    for indent in XPath('./w:ind')(parent):
        l, lc = get(indent, 'w:left'), get(indent, 'w:leftChars')
        pl = simple_float(lc, 0.01) if lc is not None else simple_float(l, 0.05) if l is not None else None
        if pl is not None:
            padding_left = '%.3g%s' % (pl, 'em' if lc is not None else 'pt')

        r, rc = get(indent, 'w:right'), get(indent, 'w:rightChars')
        pr = simple_float(rc, 0.01) if rc is not None else simple_float(r, 0.05) if r is not None else None
        if pr is not None:
            padding_right = '%.3g%s' % (pr, 'em' if rc is not None else 'pt')

        h, hc = get(indent, 'w:hanging'), get(indent, 'w:hangingChars')
        fl, flc = get(indent, 'w:firstLine'), get(indent, 'w:firstLineChars')
        h = h if h is None else '-'+h
        hc = hc if hc is None else '-'+hc
        ti = (simple_float(hc, 0.01) if hc is not None else simple_float(h, 0.05) if h is not None else
              simple_float(flc, 0.01) if flc is not None else simple_float(fl, 0.05) if fl is not None else None)
        if ti is not None:
            text_indent = '%.3g%s' % (ti, 'em' if hc is not None or (h is None and flc is not None) else 'pt')

    setattr(dest, 'margin_left', padding_left)
    setattr(dest, 'margin_right', padding_right)
    setattr(dest, 'text_indent', text_indent)

def read_justification(parent, dest):
    ans = inherit
    for jc in XPath('./w:jc[@w:val]')(parent):
        val = get(jc, 'w:val')
        if not val:
            continue
        if val in {'both', 'distribute'} or 'thai' in val or 'kashida' in val:
            ans = 'justify'
        if val in {'left', 'center', 'right',}:
            ans = val
    setattr(dest, 'text_align', ans)

def read_spacing(parent, dest):
    padding_top = padding_bottom = line_height = inherit
    for s in XPath('./w:spacing')(parent):
        a, al, aa = get(s, 'w:after'), get(s, 'w:afterLines'), get(s, 'w:afterAutospacing')
        pb = None if aa in {'on', '1', 'true'} else simple_float(al, 0.02) if al is not None else simple_float(a, 0.05) if a is not None else None
        if pb is not None:
            padding_bottom = '%.3g%s' % (pb, 'ex' if al is not None else 'pt')

        b, bl, bb = get(s, 'w:before'), get(s, 'w:beforeLines'), get(s, 'w:beforeAutospacing')
        pt = None if bb in {'on', '1', 'true'} else simple_float(bl, 0.02) if bl is not None else simple_float(b, 0.05) if b is not None else None
        if pt is not None:
            padding_top = '%.3g%s' % (pt, 'ex' if bl is not None else 'pt')

        l, lr = get(s, 'w:line'), get(s, 'w:lineRule', 'auto')
        if l is not None:
            lh = simple_float(l, 0.05) if lr in {'exact', 'atLeast'} else simple_float(l, 1/240.0)
            line_height = '%.3g%s' % (lh, 'pt' if lr in {'exact', 'atLeast'} else '')

    setattr(dest, 'margin_top', padding_top)
    setattr(dest, 'margin_bottom', padding_bottom)
    setattr(dest, 'line_height', line_height)

def read_direction(parent, dest):
    ans = inherit
    for jc in XPath('./w:textFlow[@w:val]')(parent):
        val = get(jc, 'w:val')
        if not val:
            continue
        if 'rl' in val.lower():
            ans = 'rtl'
    setattr(dest, 'direction', ans)

def read_shd(parent, dest):
    ans = inherit
    for shd in XPath('./w:shd[@w:fill]')(parent):
        val = get(shd, 'w:fill')
        if val:
            ans = simple_color(val, auto='transparent')
    setattr(dest, 'background_color', ans)

def read_numbering(parent, dest):
    lvl = num_id = None
    for np in XPath('./w:numPr')(parent):
        for ilvl in XPath('./w:ilvl[@w:val]')(np):
            try:
                lvl = int(get(ilvl, 'w:val'))
            except (ValueError, TypeError):
                pass
        for num in XPath('./w:numId[@w:val]')(np):
            num_id = get(num, 'w:val')
    val = (num_id, lvl) if num_id is not None or lvl is not None else inherit
    setattr(dest, 'numbering', val)

class Frame(object):

    all_attributes = ('drop_cap', 'h', 'w', 'h_anchor', 'h_rule', 'v_anchor', 'wrap',
                      'h_space', 'v_space', 'lines', 'x_align', 'y_align', 'x', 'y')

    def __init__(self, fp):
        self.drop_cap = get(fp, 'w:dropCap', 'none')
        try:
            self.h = int(get(fp, 'w:h'))/20
        except (ValueError, TypeError):
            self.h = 0
        try:
            self.w = int(get(fp, 'w:w'))/20
        except (ValueError, TypeError):
            self.w = None
        try:
            self.x = int(get(fp, 'w:x'))/20
        except (ValueError, TypeError):
            self.x = 0
        try:
            self.y = int(get(fp, 'w:y'))/20
        except (ValueError, TypeError):
            self.y = 0

        self.h_anchor = get(fp, 'w:hAnchor', 'page')
        self.h_rule = get(fp, 'w:hRule', 'auto')
        self.v_anchor = get(fp, 'w:vAnchor', 'page')
        self.wrap = get(fp, 'w:wrap', 'around')
        self.x_align = get(fp, 'w:xAlign')
        self.y_align = get(fp, 'w:yAlign')

        try:
            self.h_space = int(get(fp, 'w:hSpace'))/20
        except (ValueError, TypeError):
            self.h_space = 0
        try:
            self.v_space = int(get(fp, 'w:vSpace'))/20
        except (ValueError, TypeError):
            self.v_space = 0
        try:
            self.lines = int(get(fp, 'w:lines'))
        except (ValueError, TypeError):
            self.lines = 1

    def css(self, page):
        is_dropcap = self.drop_cap in {'drop', 'margin'}
        ans = {'overflow': 'hidden'}

        if is_dropcap:
            ans['float'] = 'left'
            ans['margin'] = '0'
            ans['padding-right'] = '0.2em'
        else:
            if self.h_rule != 'auto':
                t = 'min-height' if self.h_rule == 'atLeast' else 'height'
                ans[t] = '%.3gpt' % self.h
            if self.w is not None:
                ans['width'] = '%.3gpt' % self.w
            ans['padding-top'] = ans['padding-bottom'] = '%.3gpt' % self.v_space
            if self.wrap not in {None, 'none'}:
                ans['padding-left'] = ans['padding-right'] = '%.3gpt' % self.h_space
                if self.x_align is None:
                    fl = 'left' if self.x/page.width < 0.5 else 'right'
                else:
                    fl = 'right' if self.x_align == 'right' else 'left'
                ans['float'] = fl
        return ans

    def __eq__(self, other):
        for x in self.all_attributes:
            if getattr(other, x, inherit) != getattr(self, x):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

def read_frame(parent, dest):
    ans = inherit
    for fp in XPath('./w:framePr')(parent):
        ans = Frame(fp)
    setattr(dest, 'frame', ans)

# }}}

class ParagraphStyle(object):

    all_properties = (
        'adjustRightInd', 'autoSpaceDE', 'autoSpaceDN', 'bidi',
        'contextualSpacing', 'keepLines', 'keepNext', 'mirrorIndents',
        'pageBreakBefore', 'snapToGrid', 'suppressLineNumbers',
        'suppressOverlap', 'topLinePunct', 'widowControl', 'wordWrap',

        # Border margins padding
        'border_left_width', 'border_left_style', 'border_left_color', 'padding_left',
        'border_top_width', 'border_top_style', 'border_top_color', 'padding_top',
        'border_right_width', 'border_right_style', 'border_right_color', 'padding_right',
        'border_bottom_width', 'border_bottom_style', 'border_bottom_color', 'padding_bottom',
        'margin_left', 'margin_top', 'margin_right', 'margin_bottom',

        # Misc.
        'text_indent', 'text_align', 'line_height', 'direction', 'background_color',
        'numbering', 'font_family', 'font_size', 'color', 'frame',
    )

    def __init__(self, pPr=None):
        self.linked_style = None
        if pPr is None:
            for p in self.all_properties:
                setattr(self, p, inherit)
        else:
            for p in (
                'adjustRightInd', 'autoSpaceDE', 'autoSpaceDN', 'bidi',
                'contextualSpacing', 'keepLines', 'keepNext', 'mirrorIndents',
                'pageBreakBefore', 'snapToGrid', 'suppressLineNumbers',
                'suppressOverlap', 'topLinePunct', 'widowControl', 'wordWrap',
            ):
                setattr(self, p, binary_property(pPr, p))

            for x in ('border', 'indent', 'justification', 'spacing', 'direction', 'shd', 'numbering', 'frame'):
                f = globals()['read_%s' % x]
                f(pPr, self)

            for s in XPath('./w:pStyle[@w:val]')(pPr):
                self.linked_style = get(s, 'w:val')

            self.font_family = self.font_size = self.color = inherit

        self._css = None

    def update(self, other):
        for prop in self.all_properties:
            nval = getattr(other, prop)
            if nval is not inherit:
                setattr(self, prop, nval)
        if other.linked_style is not None:
            self.linked_style = other.linked_style

    def resolve_based_on(self, parent):
        for p in self.all_properties:
            val = getattr(self, p)
            if val is inherit:
                setattr(self, p, getattr(parent, p))

    @property
    def css(self):
        if self._css is None:
            self._css = c = OrderedDict()
            if self.keepLines is True:
                c['page-break-inside'] = 'avoid'
            if self.pageBreakBefore is True:
                c['page-break-before'] = 'always'
            if self.keepNext is True:
                c['page-break-after'] = 'avoid'
            for edge in ('left', 'top', 'right', 'bottom'):
                val = getattr(self, 'border_%s_width' % edge)
                if val is not inherit:
                    c['border-left-width'] = '%.3gpt' % val
                for x in ('style', 'color'):
                    val = getattr(self, 'border_%s_%s' % (edge, x))
                    if val is not inherit:
                        c['border-%s-%s' % (edge, x)] = val
                val = getattr(self, 'padding_%s' % edge)
                if val is not inherit:
                    c['padding-%s' % edge] = '%.3gpt' % val
                val = getattr(self, 'margin_%s' % edge)
                if val is not inherit:
                    c['margin-%s' % edge] = val

            if self.line_height not in {inherit, '1'}:
                c['line-height'] = self.line_height

            for x in ('text_indent', 'text_align', 'background_color', 'font_family', 'font_size', 'color'):
                val = getattr(self, x)
                if val is not inherit:
                    if x == 'font_size':
                        val = '%.3gpt' % val
                    c[x.replace('_', '-')] = val

        return self._css




# -*- coding: utf-8 -*-

from __future__ import (unicode_literals, division, absolute_import, print_function)
store_version = 2  # Needed for dynamic plugin loading

__license__ = 'GPL 3'
__copyright__ = '2011, John Schember <john@nachtimwald.com>'
__docformat__ = 'restructuredtext en'

import urllib
from contextlib import closing

from lxml import html

from PyQt5.Qt import QUrl

from calibre import browser, url_slash_cleaner
from calibre.gui2 import open_url
from calibre.gui2.store import StorePlugin
from calibre.gui2.store.basic_config import BasicStoreConfig
from calibre.gui2.store.search_result import SearchResult
from calibre.gui2.store.web_store_dialog import WebStoreDialog

class GoogleBooksStore(BasicStoreConfig, StorePlugin):

    def open(self, parent=None, detail_item=None, external=False):
        url = 'http://books.google.com/books'
        if external or self.config.get('open_external', False):
            open_url(QUrl(url_slash_cleaner(detail_item if detail_item else url)))
        else:
            d = WebStoreDialog(self.gui, url, parent, detail_item)
            d.setWindowTitle(self.name)
            d.set_tags(self.config.get('tags', ''))
            d.exec_()

    def search(self, query, max_results=10, timeout=60):
        url = 'http://www.google.com/search?tbm=bks&q=' + urllib.quote_plus(query)

        br = browser()

        counter = max_results
        with closing(br.open(url, timeout=timeout)) as f:
            doc = html.fromstring(f.read())
            for data in doc.xpath('//ol/li'):
                if counter <= 0:
                    break

                id = ''.join(data.xpath('.//h3/a/@href'))
                if not id:
                    continue

                title = ''.join(data.xpath('.//h3/a//text()'))
                authors = data.xpath('.//span[contains(@class, "f")]//a//text()')
                while authors and authors[-1].strip().lower() in ('preview', 'read', 'more editions'):
                    authors = authors[:-1]
                if not authors:
                    continue
                author = ', '.join(authors)

                counter -= 1

                s = SearchResult()
                s.title = title.strip()
                s.author = author.strip()
                s.detail_item = id.strip()
                s.drm = SearchResult.DRM_UNKNOWN

                yield s

    def get_details(self, search_result, timeout):
        br = browser()
        with closing(br.open(search_result.detail_item, timeout=timeout)) as nf:
            doc = html.fromstring(nf.read())

            search_result.cover_url = ''.join(doc.xpath('//div[@class="sidebarcover"]//img/@src'))

            # Try to get the set price.
            price = ''.join(doc.xpath('//div[@id="gb-get-book-container"]//a/text()'))
            if 'read' in price.lower():
                price = 'Unknown'
            elif 'free' in price.lower() or not price.strip():
                price = '$0.00'
            elif '-' in price:
                a, b, price = price.partition(' - ')
            search_result.price = price.strip()

            search_result.formats = ', '.join(doc.xpath('//div[contains(@class, "download-panel-div")]//a/text()')).upper()
            if not search_result.formats:
                search_result.formats = _('Unknown')

        return True




#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
from __future__ import (unicode_literals, division, absolute_import,
                        print_function)

__license__   = 'GPL v3'
__copyright__ = '2011, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

from calibre.gui2.complete2 import EditWithComplete
from calibre.utils.localization import lang_map
from calibre.utils.icu import sort_key, lower

class LanguagesEdit(EditWithComplete):

    def __init__(self, parent=None, db=None):
        EditWithComplete.__init__(self, parent)

        self.setSizeAdjustPolicy(self.AdjustToMinimumContentsLengthWithIcon)
        self.setMinimumContentsLength(20)
        self._lang_map = lang_map()
        self.names_with_commas = [x for x in self._lang_map.itervalues() if ',' in x]
        self.comma_map = {k:k.replace(',', '|') for k in self.names_with_commas}
        self.comma_rmap = {v:k for k, v in self.comma_map.iteritems()}
        self._rmap = {lower(v):k for k,v in self._lang_map.iteritems()}
        self.init_langs(db)

    def init_langs(self, db):
        if db is not None:
            pmap = {self._lang_map.get(x[1], x[1]):1 for x in
                    db.get_languages_with_ids()}
            all_items = sorted(self._lang_map.itervalues(),
                key=lambda x: (-pmap.get(x, 0), sort_key(x)))
        else:
            all_items = sorted(self._lang_map.itervalues(),
                key=lambda x: sort_key(x))
        self.update_items_cache(all_items)

    @property
    def vals(self):
        raw = unicode(self.lineEdit().text())
        for k, v in self.comma_map.iteritems():
            raw = raw.replace(k, v)
        parts = [x.strip() for x in raw.split(',')]
        return [self.comma_rmap.get(x, x) for x in parts]

    @dynamic_property
    def lang_codes(self):

        def fget(self):
            vals = self.vals
            ans = []
            for name in vals:
                if name:
                    code = self._rmap.get(lower(name), None)
                    if code is not None:
                        ans.append(code)
            return ans

        def fset(self, lang_codes):
            self.set_lang_codes(lang_codes, allow_undo=False)

        return property(fget=fget, fset=fset)

    def set_lang_codes(self, lang_codes, allow_undo=True):
        ans = []
        for lc in lang_codes:
            name = self._lang_map.get(lc, None)
            if name is not None:
                ans.append(name)
        ans = ', '.join(ans)

        if allow_undo:
            orig, self.disable_popup = self.disable_popup, True
            try:
                self.lineEdit().selectAll(), self.lineEdit().insert(ans)
            finally:
                self.disable_popup = orig
        else:
            self.setEditText(ans)

    def validate(self):
        vals = self.vals
        bad = []
        for name in vals:
            if name:
                code = self._rmap.get(lower(name), None)
                if code is None:
                    bad.append(name)
        return bad




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/preferences/create_custom_column.ui'
#
# Created: Fri Aug  8 18:32:27 2014
#      by: PyQt5 UI code generator 5.3.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_QCreateCustomColumn(object):
    def setupUi(self, QCreateCustomColumn):
        QCreateCustomColumn.setObjectName("QCreateCustomColumn")
        QCreateCustomColumn.setWindowModality(QtCore.Qt.ApplicationModal)
        QCreateCustomColumn.resize(831, 344)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(QCreateCustomColumn.sizePolicy().hasHeightForWidth())
        QCreateCustomColumn.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(I("column.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        QCreateCustomColumn.setWindowIcon(icon)
        self.verticalLayout = QtWidgets.QVBoxLayout(QCreateCustomColumn)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout_2.setContentsMargins(5, 5, 5, 5)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.label_2 = QtWidgets.QLabel(QCreateCustomColumn)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 0, 0, 1, 1)
        self.label = QtWidgets.QLabel(QCreateCustomColumn)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)
        self.column_name_box = QtWidgets.QLineEdit(QCreateCustomColumn)
        self.column_name_box.setMinimumSize(QtCore.QSize(20, 0))
        self.column_name_box.setObjectName("column_name_box")
        self.gridLayout.addWidget(self.column_name_box, 0, 2, 1, 1)
        self.column_heading_box = QtWidgets.QLineEdit(QCreateCustomColumn)
        self.column_heading_box.setObjectName("column_heading_box")
        self.gridLayout.addWidget(self.column_heading_box, 1, 2, 1, 1)
        self.label_3 = QtWidgets.QLabel(QCreateCustomColumn)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.column_type_box = QtWidgets.QComboBox(QCreateCustomColumn)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.column_type_box.sizePolicy().hasHeightForWidth())
        self.column_type_box.setSizePolicy(sizePolicy)
        self.column_type_box.setMinimumSize(QtCore.QSize(70, 0))
        self.column_type_box.setObjectName("column_type_box")
        self.horizontalLayout_3.addWidget(self.column_type_box)
        self.use_decorations = QtWidgets.QCheckBox(QCreateCustomColumn)
        self.use_decorations.setObjectName("use_decorations")
        self.horizontalLayout_3.addWidget(self.use_decorations)
        self.is_names = QtWidgets.QCheckBox(QCreateCustomColumn)
        self.is_names.setObjectName("is_names")
        self.horizontalLayout_3.addWidget(self.is_names)
        spacerItem = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout_3, 2, 2, 1, 1)
        self.horizontalLayout_31 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_31.setObjectName("horizontalLayout_31")
        self.date_format_box = QtWidgets.QLineEdit(QCreateCustomColumn)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.date_format_box.sizePolicy().hasHeightForWidth())
        self.date_format_box.setSizePolicy(sizePolicy)
        self.date_format_box.setObjectName("date_format_box")
        self.horizontalLayout_31.addWidget(self.date_format_box)
        self.number_format_box = QtWidgets.QLineEdit(QCreateCustomColumn)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.number_format_box.sizePolicy().hasHeightForWidth())
        self.number_format_box.setSizePolicy(sizePolicy)
        self.number_format_box.setObjectName("number_format_box")
        self.horizontalLayout_31.addWidget(self.number_format_box)
        self.date_format_default_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.date_format_default_label.setObjectName("date_format_default_label")
        self.horizontalLayout_31.addWidget(self.date_format_default_label)
        self.number_format_default_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.number_format_default_label.setOpenExternalLinks(True)
        self.number_format_default_label.setObjectName("number_format_default_label")
        self.horizontalLayout_31.addWidget(self.number_format_default_label)
        self.gridLayout.addLayout(self.horizontalLayout_31, 4, 2, 1, 1)
        self.date_format_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.date_format_label.setObjectName("date_format_label")
        self.gridLayout.addWidget(self.date_format_label, 4, 0, 1, 1)
        self.number_format_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.number_format_label.setObjectName("number_format_label")
        self.gridLayout.addWidget(self.number_format_label, 4, 0, 1, 1)
        self.composite_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.composite_label.setObjectName("composite_label")
        self.gridLayout.addWidget(self.composite_label, 5, 0, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.composite_box = QtWidgets.QLineEdit(QCreateCustomColumn)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.composite_box.sizePolicy().hasHeightForWidth())
        self.composite_box.setSizePolicy(sizePolicy)
        self.composite_box.setObjectName("composite_box")
        self.horizontalLayout_4.addWidget(self.composite_box)
        self.composite_default_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.composite_default_label.setObjectName("composite_default_label")
        self.horizontalLayout_4.addWidget(self.composite_default_label)
        self.gridLayout.addLayout(self.horizontalLayout_4, 5, 2, 1, 1)
        self.composite_layout = QtWidgets.QHBoxLayout()
        self.composite_layout.setObjectName("composite_layout")
        self.composite_sort_by_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.composite_sort_by_label.setObjectName("composite_sort_by_label")
        self.composite_layout.addWidget(self.composite_sort_by_label)
        self.composite_sort_by = QtWidgets.QComboBox(QCreateCustomColumn)
        self.composite_sort_by.setObjectName("composite_sort_by")
        self.composite_layout.addWidget(self.composite_sort_by)
        self.composite_make_category = QtWidgets.QCheckBox(QCreateCustomColumn)
        self.composite_make_category.setObjectName("composite_make_category")
        self.composite_layout.addWidget(self.composite_make_category)
        self.composite_contains_html = QtWidgets.QCheckBox(QCreateCustomColumn)
        self.composite_contains_html.setObjectName("composite_contains_html")
        self.composite_layout.addWidget(self.composite_contains_html)
        spacerItem1 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.composite_layout.addItem(spacerItem1)
        self.gridLayout.addLayout(self.composite_layout, 6, 2, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem2, 11, 0, 1, 4)
        self.enum_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.enum_label.setObjectName("enum_label")
        self.gridLayout.addWidget(self.enum_label, 6, 0, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QGridLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.enum_box = QtWidgets.QLineEdit(QCreateCustomColumn)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.enum_box.sizePolicy().hasHeightForWidth())
        self.enum_box.setSizePolicy(sizePolicy)
        self.enum_box.setObjectName("enum_box")
        self.horizontalLayout_2.addWidget(self.enum_box, 0, 0, 1, 1)
        self.enum_default_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.enum_default_label.setObjectName("enum_default_label")
        self.horizontalLayout_2.addWidget(self.enum_default_label, 0, 1, 1, 1)
        self.enum_colors = QtWidgets.QLineEdit(QCreateCustomColumn)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.enum_colors.sizePolicy().hasHeightForWidth())
        self.enum_colors.setSizePolicy(sizePolicy)
        self.enum_colors.setObjectName("enum_colors")
        self.horizontalLayout_2.addWidget(self.enum_colors, 1, 0, 1, 1)
        self.enum_colors_label = QtWidgets.QLabel(QCreateCustomColumn)
        self.enum_colors_label.setObjectName("enum_colors_label")
        self.horizontalLayout_2.addWidget(self.enum_colors_label, 1, 1, 1, 1)
        self.gridLayout.addLayout(self.horizontalLayout_2, 6, 2, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout, 5, 0, 1, 1)
        self.button_box = QtWidgets.QDialogButtonBox(QCreateCustomColumn)
        self.button_box.setOrientation(QtCore.Qt.Horizontal)
        self.button_box.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.button_box.setCenterButtons(True)
        self.button_box.setObjectName("button_box")
        self.gridLayout_2.addWidget(self.button_box, 14, 0, 1, 1)
        self.heading_label = QtWidgets.QLabel(QCreateCustomColumn)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.heading_label.setFont(font)
        self.heading_label.setText("")
        self.heading_label.setObjectName("heading_label")
        self.gridLayout_2.addWidget(self.heading_label, 1, 0, 1, 1)
        self.shortcuts = QtWidgets.QLabel(QCreateCustomColumn)
        self.shortcuts.setText("")
        self.shortcuts.setWordWrap(True)
        self.shortcuts.setObjectName("shortcuts")
        self.gridLayout_2.addWidget(self.shortcuts, 3, 0, 1, 1)
        self.line = QtWidgets.QFrame(QCreateCustomColumn)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout_2.addWidget(self.line, 2, 0, 1, 1)
        self.line_2 = QtWidgets.QFrame(QCreateCustomColumn)
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout_2.addWidget(self.line_2, 4, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout_2)
        self.label_2.setBuddy(self.column_name_box)
        self.label.setBuddy(self.column_heading_box)
        self.label_3.setBuddy(self.column_type_box)
        self.date_format_label.setBuddy(self.date_format_box)
        self.number_format_label.setBuddy(self.number_format_box)
        self.composite_label.setBuddy(self.composite_box)
        self.composite_sort_by_label.setBuddy(self.composite_sort_by)
        self.enum_label.setBuddy(self.enum_box)

        self.retranslateUi(QCreateCustomColumn)
        QtCore.QMetaObject.connectSlotsByName(QCreateCustomColumn)
        QCreateCustomColumn.setTabOrder(self.column_name_box, self.column_heading_box)
        QCreateCustomColumn.setTabOrder(self.column_heading_box, self.column_type_box)
        QCreateCustomColumn.setTabOrder(self.column_type_box, self.date_format_box)
        QCreateCustomColumn.setTabOrder(self.date_format_box, self.composite_box)
        QCreateCustomColumn.setTabOrder(self.composite_box, self.button_box)

    def retranslateUi(self, QCreateCustomColumn):

        self.label_2.setText(_("&Lookup name"))
        self.label.setText(_("Column &heading"))
        self.column_name_box.setToolTip(_("Used for searching the column. Must contain only digits and lower case letters."))
        self.column_heading_box.setToolTip(_("Column heading in the library view and category name in the tag browser"))
        self.label_3.setText(_("&Column type"))
        self.column_type_box.setToolTip(_("What kind of information will be kept in the column."))
        self.use_decorations.setToolTip(_("Show check marks in the GUI. Values of \'yes\', \'checked\', and \'true\'\n"
"will show a green check. Values of \'no\', \'unchecked\', and \'false\' will show a red X.\n"
"Everything else will show nothing."))
        self.use_decorations.setText(_("Show checkmarks"))
        self.is_names.setToolTip(_("Check this box if this column contains names, like the authors column."))
        self.is_names.setText(_("Contains names"))
        self.date_format_box.setToolTip(_("<p>Date format. Use 1-4 \'d\'s for day, 1-4 \'M\'s for month, and 2 or 4 \'y\'s for year.</p>\n"
"<p>For example:\n"
"<ul>\n"
"<li> ddd, d MMM yyyy gives Mon, 5 Jan 2010<li>\n"
"<li>dd MMMM yy gives 05 January 10</li>\n"
"</ul> "))
        self.date_format_default_label.setToolTip(_("Use MMM yyyy for month + year, yyyy for year only"))
        self.date_format_default_label.setText(_("Default: dd MMM yyyy."))
        self.number_format_default_label.setToolTip(_("<p>The format specifier must begin with <code>{0:</code>\n"
"and end with <code>}</code> You can have text before and after the format specifier.\n"
"            "))
        self.number_format_default_label.setText(_("<p>Default: Not formatted. For format language details see <a href=\"http://docs.python.org/library/string.html#format-string-syntax\">the python documentation</a>"))
        self.date_format_label.setText(_("Format for &dates"))
        self.number_format_label.setText(_("Format for &numbers"))
        self.composite_label.setText(_("&Template"))
        self.composite_box.setToolTip(_("Field template. Uses the same syntax as save templates."))
        self.composite_default_label.setToolTip(_("Similar to save templates. For example, {title} {isbn}"))
        self.composite_default_label.setText(_("Default: (nothing)"))
        self.composite_sort_by_label.setText(_("&Sort/search column by"))
        self.composite_sort_by.setToolTip(_("How this column should handled in the GUI when sorting and searching"))
        self.composite_make_category.setToolTip(_("If checked, this column will appear in the tags browser as a category"))
        self.composite_make_category.setText(_("Show in tags browser"))
        self.composite_contains_html.setText(_("Show as HTML in book details"))
        self.enum_label.setText(_("Values"))
        self.enum_box.setToolTip(_("A comma-separated list of permitted values. The empty value is always\n"
"included, and is the default. For example, the list \'one,two,three\' has\n"
"four values, the first of them being the empty value."))
        self.enum_default_label.setToolTip(_("The empty string is always the first value"))
        self.enum_default_label.setText(_("Values"))
        self.enum_colors.setToolTip(_("A list of color names to use when displaying an item. The\n"
"list must be empty or contain a color for each value."))
        self.enum_colors_label.setText(_("Colors"))





#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai

__license__   = 'GPL v3'
__copyright__ = '2010, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import os, cPickle, signal, time
from Queue import Queue, Empty
from multiprocessing.connection import Listener, arbitrary_address
from binascii import hexlify

from PyQt5.Qt import QThread, pyqtSignal

from calibre.utils.pyconsole import Process, iswindows, POLL_TIMEOUT

class Controller(QThread):

    # show_error(is_syntax_error, traceback, self)
    show_error = pyqtSignal(object, object, object)
    # write_output(unicode_object, stdout or stderr, self)
    write_output = pyqtSignal(object, object, object)
    # Indicates interpreter has finished evaluating current command
    interpreter_done = pyqtSignal(object, object)
    # interpreter_died(self, returncode or None if no return code available)
    interpreter_died = pyqtSignal(object, object)

    def __init__(self, parent):
        QThread.__init__(self, parent)
        self.keep_going = True
        self.current_command = None

        self.out_queue = Queue()
        self.address = arbitrary_address('AF_PIPE' if iswindows else 'AF_UNIX')
        self.auth_key = os.urandom(32)
        if iswindows and self.address[1] == ':':
            self.address = self.address[2:]
        self.listener = Listener(address=self.address,
                authkey=self.auth_key, backlog=4)

        self.env = {
            'CALIBRE_SIMPLE_WORKER':
                'calibre.utils.pyconsole.interpreter:main',
            'CALIBRE_WORKER_ADDRESS':
                    hexlify(cPickle.dumps(self.listener.address, -1)),
            'CALIBRE_WORKER_KEY': hexlify(self.auth_key)
        }
        self.process = Process(self.env)
        self.output_file_buf = self.process(redirect_output=False)
        self.conn = self.listener.accept()
        self.start()

    def run(self):
        while self.keep_going and self.is_alive:
            try:
                self.communicate()
            except KeyboardInterrupt:
                pass
            except EOFError:
                break
        self.interpreter_died.emit(self, self.returncode)
        try:
            self.listener.close()
        except:
            pass

    def communicate(self):
        if self.conn.poll(POLL_TIMEOUT):
            self.dispatch_incoming_message(self.conn.recv())
        try:
            obj = self.out_queue.get_nowait()
        except Empty:
            pass
        else:
            try:
                self.conn.send(obj)
            except:
                raise EOFError('controller failed to send')

    def dispatch_incoming_message(self, obj):
        try:
            cmd, data = obj
        except:
            print 'Controller received invalid message'
            print repr(obj)
            return
        if cmd in ('stdout', 'stderr'):
            self.write_output.emit(data, cmd, self)
        elif cmd == 'syntaxerror':
            self.show_error.emit(True, data, self)
        elif cmd == 'traceback':
            self.show_error.emit(False, data, self)
        elif cmd == 'done':
            self.current_command = None
            self.interpreter_done.emit(self, data)

    def runsource(self, cmd):
        self.current_command = cmd
        self.out_queue.put(('run', cmd))

    def __nonzero__(self):
        return self.process.is_alive

    @property
    def returncode(self):
        return self.process.returncode

    def interrupt(self):
        if hasattr(signal, 'SIGINT'):
            os.kill(self.process.pid, signal.SIGINT)
        elif hasattr(signal, 'CTRL_C_EVENT'):
            os.kill(self.process.pid, signal.CTRL_C_EVENT)

    @property
    def is_alive(self):
        return self.process.is_alive

    def kill(self):
        self.out_queue.put(('quit', 0))
        t = 0
        while self.is_alive and t < 10:
            time.sleep(0.1)
        self.process.kill()
        self.keep_going = False




#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
from __future__ import (unicode_literals, division, absolute_import,
                        print_function)

__license__   = 'GPL v3'
__copyright__ = '2009, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import re, codecs

ENCODING_PATS = [
    # XML declaration
    re.compile(r'<\?[^<>]+encoding\s*=\s*[\'"](.*?)[\'"][^<>]*>', re.IGNORECASE),
    # HTML 5 charset
    re.compile(r'''<meta\s+charset=['"]([-_a-z0-9]+)['"][^<>]*>(?:\s*</meta>){0,1}''', re.IGNORECASE),
    # HTML 4 Pragma directive
    re.compile(r'''<meta\s+?[^<>]*?content\s*=\s*['"][^'"]*?charset=([-_a-z0-9]+)[^'"]*?['"][^<>]*>(?:\s*</meta>){0,1}''', re.IGNORECASE),
]
ENTITY_PATTERN = re.compile(r'&(\S+?);')

def strip_encoding_declarations(raw, limit=50*1024):
    prefix = raw[:limit]
    suffix = raw[limit:]
    for pat in ENCODING_PATS:
        prefix = pat.sub('', prefix)
    raw = prefix + suffix
    return raw

def replace_encoding_declarations(raw, enc='utf-8', limit=50*1024):
    prefix = raw[:limit]
    suffix = raw[limit:]
    changed = [False]
    def sub(m):
        ans = m.group()
        if m.group(1).lower() != enc.lower():
            changed[0] = True
            start, end = m.start(1) - m.start(0), m.end(1) - m.end(0)
            ans = ans[:start] + enc + ans[end:]
        return ans

    for pat in ENCODING_PATS:
        prefix = pat.sub(sub, prefix)
    raw = prefix + suffix
    return raw, changed[0]

def find_declared_encoding(raw, limit=50*1024):
    prefix = raw[:limit]
    for pat in ENCODING_PATS:
        m = pat.search(prefix)
        if m is not None:
            return m.group(1)

def substitute_entites(raw):
    from calibre import xml_entity_to_unicode
    return ENTITY_PATTERN.sub(xml_entity_to_unicode, raw)

_CHARSET_ALIASES = {"macintosh" : "mac-roman",
                        "x-sjis" : "shift-jis"}

def detect(*args, **kwargs):
    from chardet import detect
    return detect(*args, **kwargs)

def force_encoding(raw, verbose, assume_utf8=False):
    from calibre.constants import preferred_encoding

    try:
        chardet = detect(raw[:1024*50])
    except:
        chardet = {'encoding':preferred_encoding, 'confidence':0}
    encoding = chardet['encoding']
    if chardet['confidence'] < 1 and assume_utf8:
        encoding = 'utf-8'
    if chardet['confidence'] < 1 and verbose:
        print('WARNING: Encoding detection confidence %d%%'%(
            chardet['confidence']*100))
    if not encoding:
        encoding = preferred_encoding
    encoding = encoding.lower()
    encoding = _CHARSET_ALIASES.get(encoding, encoding)
    if encoding == 'ascii':
        encoding = 'utf-8'
    return encoding

def detect_xml_encoding(raw, verbose=False, assume_utf8=False):
    if not raw or isinstance(raw, unicode):
        return raw, None
    for x in ('utf8', 'utf-16-le', 'utf-16-be'):
        bom = getattr(codecs, 'BOM_'+x.upper().replace('-16', '16').replace(
            '-', '_'))
        if raw.startswith(bom):
            return raw[len(bom):], x
    encoding = None
    for pat in ENCODING_PATS:
        match = pat.search(raw)
        if match:
            encoding = match.group(1)
            break
    if encoding is None:
        encoding = force_encoding(raw, verbose, assume_utf8=assume_utf8)
    if encoding.lower().strip() == 'macintosh':
        encoding = 'mac-roman'
    if encoding.lower().replace('_', '-').strip() in (
            'gb2312', 'chinese', 'csiso58gb231280', 'euc-cn', 'euccn',
            'eucgb2312-cn', 'gb2312-1980', 'gb2312-80', 'iso-ir-58'):
        # Microsoft Word exports to HTML with encoding incorrectly set to
        # gb2312 instead of gbk. gbk is a superset of gb2312, anyway.
        encoding = 'gbk'
    try:
        codecs.lookup(encoding)
    except LookupError:
        encoding = 'utf-8'

    return raw, encoding

def xml_to_unicode(raw, verbose=False, strip_encoding_pats=False,
                   resolve_entities=False, assume_utf8=False):
    '''
    Force conversion of byte string to unicode. Tries to look for XML/HTML
    encoding declaration first, if not found uses the chardet library and
    prints a warning if detection confidence is < 100%
    @return: (unicode, encoding used)
    '''
    if not raw:
        return '', None
    raw, encoding = detect_xml_encoding(raw, verbose=verbose,
            assume_utf8=assume_utf8)
    if not isinstance(raw, unicode):
        raw = raw.decode(encoding, 'replace')

    if strip_encoding_pats:
        raw = strip_encoding_declarations(raw)
    if resolve_entities:
        raw = substitute_entites(raw)

    return raw, encoding



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/dialogs/user_profiles.ui'
#
# Created: Sat Aug  9 08:52:10 2014
#      by: PyQt5 UI code generator 5.3.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(738, 640)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(I("user_profile.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        Dialog.setWindowIcon(icon)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.scrollArea = QtWidgets.QScrollArea(Dialog)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setLineWidth(0)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 726, 595))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.central_widget = QtWidgets.QWidget(self.scrollAreaWidgetContents)
        self.central_widget.setMinimumSize(QtCore.QSize(580, 550))
        self.central_widget.setObjectName("central_widget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.central_widget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.groupBox = QtWidgets.QGroupBox(self.central_widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.available_profiles = QtWidgets.QListView(self.groupBox)
        self.available_profiles.setObjectName("available_profiles")
        self.verticalLayout_2.addWidget(self.available_profiles)
        self.add_profile_button = QtWidgets.QPushButton(self.groupBox)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(I("plus.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_profile_button.setIcon(icon1)
        self.add_profile_button.setObjectName("add_profile_button")
        self.verticalLayout_2.addWidget(self.add_profile_button)
        self.remove_profile_button = QtWidgets.QPushButton(self.groupBox)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(I("list_remove.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.remove_profile_button.setIcon(icon2)
        self.remove_profile_button.setObjectName("remove_profile_button")
        self.verticalLayout_2.addWidget(self.remove_profile_button)
        self.share_button = QtWidgets.QPushButton(self.groupBox)
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(I("forward.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.share_button.setIcon(icon3)
        self.share_button.setObjectName("share_button")
        self.verticalLayout_2.addWidget(self.share_button)
        self.show_recipe_files_button = QtWidgets.QPushButton(self.groupBox)
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(I("document_open.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.show_recipe_files_button.setIcon(icon4)
        self.show_recipe_files_button.setObjectName("show_recipe_files_button")
        self.verticalLayout_2.addWidget(self.show_recipe_files_button)
        self.builtin_recipe_button = QtWidgets.QPushButton(self.groupBox)
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(I("news.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.builtin_recipe_button.setIcon(icon5)
        self.builtin_recipe_button.setObjectName("builtin_recipe_button")
        self.verticalLayout_2.addWidget(self.builtin_recipe_button)
        self.load_button = QtWidgets.QPushButton(self.groupBox)
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(I("chapters.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.load_button.setIcon(icon6)
        self.load_button.setObjectName("load_button")
        self.verticalLayout_2.addWidget(self.load_button)
        self.opml_button = QtWidgets.QPushButton(self.groupBox)
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(I("mimetypes/opml.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.opml_button.setIcon(icon7)
        self.opml_button.setObjectName("opml_button")
        self.verticalLayout_2.addWidget(self.opml_button)
        self.horizontalLayout.addWidget(self.groupBox)
        self.frame = QtWidgets.QFrame(self.central_widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(10)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout.setObjectName("verticalLayout")
        self.toggle_mode_button = QtWidgets.QPushButton(self.frame)
        self.toggle_mode_button.setObjectName("toggle_mode_button")
        self.verticalLayout.addWidget(self.toggle_mode_button)
        self.stacks = QtWidgets.QStackedWidget(self.frame)
        self.stacks.setObjectName("stacks")
        self.page = QtWidgets.QWidget()
        self.page.setObjectName("page")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.page)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label = QtWidgets.QLabel(self.page)
        self.label.setTextFormat(QtCore.Qt.RichText)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_5.addWidget(self.label)
        self.gridlayout = QtWidgets.QGridLayout()
        self.gridlayout.setObjectName("gridlayout")
        self.label_2 = QtWidgets.QLabel(self.page)
        self.label_2.setObjectName("label_2")
        self.gridlayout.addWidget(self.label_2, 0, 0, 1, 1)
        self.profile_title = EnLineEdit(self.page)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.profile_title.setFont(font)
        self.profile_title.setObjectName("profile_title")
        self.gridlayout.addWidget(self.profile_title, 0, 1, 1, 2)
        self.label_6 = QtWidgets.QLabel(self.page)
        self.label_6.setObjectName("label_6")
        self.gridlayout.addWidget(self.label_6, 2, 0, 1, 1)
        self.oldest_article = QtWidgets.QSpinBox(self.page)
        self.oldest_article.setMinimum(1)
        self.oldest_article.setMaximum(36500)
        self.oldest_article.setProperty("value", 7)
        self.oldest_article.setObjectName("oldest_article")
        self.gridlayout.addWidget(self.oldest_article, 2, 2, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.page)
        self.label_7.setObjectName("label_7")
        self.gridlayout.addWidget(self.label_7, 3, 0, 1, 1)
        self.max_articles = QtWidgets.QSpinBox(self.page)
        self.max_articles.setMinimum(5)
        self.max_articles.setMaximum(100)
        self.max_articles.setProperty("value", 10)
        self.max_articles.setObjectName("max_articles")
        self.gridlayout.addWidget(self.max_articles, 3, 2, 1, 1)
        self.verticalLayout_5.addLayout(self.gridlayout)
        self.groupBox_2 = QtWidgets.QGroupBox(self.page)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setObjectName("groupBox_2")
        self.hboxlayout = QtWidgets.QHBoxLayout(self.groupBox_2)
        self.hboxlayout.setObjectName("hboxlayout")
        self.added_feeds = BasicList(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.added_feeds.sizePolicy().hasHeightForWidth())
        self.added_feeds.setSizePolicy(sizePolicy)
        self.added_feeds.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.added_feeds.setObjectName("added_feeds")
        self.hboxlayout.addWidget(self.added_feeds)
        self.vboxlayout = QtWidgets.QVBoxLayout()
        self.vboxlayout.setObjectName("vboxlayout")
        self.up_button = QtWidgets.QToolButton(self.groupBox_2)
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap(I("arrow-up.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.up_button.setIcon(icon8)
        self.up_button.setObjectName("up_button")
        self.vboxlayout.addWidget(self.up_button)
        self.remove_feed_button = QtWidgets.QToolButton(self.groupBox_2)
        self.remove_feed_button.setIcon(icon2)
        self.remove_feed_button.setObjectName("remove_feed_button")
        self.vboxlayout.addWidget(self.remove_feed_button)
        self.down_button = QtWidgets.QToolButton(self.groupBox_2)
        icon9 = QtGui.QIcon()
        icon9.addPixmap(QtGui.QPixmap(I("arrow-down.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.down_button.setIcon(icon9)
        self.down_button.setObjectName("down_button")
        self.vboxlayout.addWidget(self.down_button)
        self.hboxlayout.addLayout(self.vboxlayout)
        self.verticalLayout_5.addWidget(self.groupBox_2)
        self.groupBox_3 = QtWidgets.QGroupBox(self.page)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridlayout1 = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridlayout1.setObjectName("gridlayout1")
        self.label_4 = QtWidgets.QLabel(self.groupBox_3)
        self.label_4.setObjectName("label_4")
        self.gridlayout1.addWidget(self.label_4, 0, 0, 1, 1)
        self.feed_title = EnLineEdit(self.groupBox_3)
        self.feed_title.setObjectName("feed_title")
        self.gridlayout1.addWidget(self.feed_title, 0, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.groupBox_3)
        self.label_5.setObjectName("label_5")
        self.gridlayout1.addWidget(self.label_5, 1, 0, 1, 1)
        self.feed_url = QtWidgets.QLineEdit(self.groupBox_3)
        self.feed_url.setObjectName("feed_url")
        self.gridlayout1.addWidget(self.feed_url, 1, 1, 1, 1)
        self.add_feed_button = QtWidgets.QPushButton(self.groupBox_3)
        self.add_feed_button.setIcon(icon1)
        self.add_feed_button.setObjectName("add_feed_button")
        self.gridlayout1.addWidget(self.add_feed_button, 2, 0, 1, 2)
        self.verticalLayout_5.addWidget(self.groupBox_3)
        self.stacks.addWidget(self.page)
        self.page_2 = QtWidgets.QWidget()
        self.page_2.setObjectName("page_2")
        self.vboxlayout1 = QtWidgets.QVBoxLayout(self.page_2)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.label_8 = QtWidgets.QLabel(self.page_2)
        self.label_8.setWordWrap(True)
        self.label_8.setOpenExternalLinks(True)
        self.label_8.setObjectName("label_8")
        self.vboxlayout1.addWidget(self.label_8)
        self.groupBox_4 = QtWidgets.QGroupBox(self.page_2)
        self.groupBox_4.setObjectName("groupBox_4")
        self.vboxlayout2 = QtWidgets.QVBoxLayout(self.groupBox_4)
        self.vboxlayout2.setObjectName("vboxlayout2")
        self.source_code = QtWidgets.QTextEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.source_code.sizePolicy().hasHeightForWidth())
        self.source_code.setSizePolicy(sizePolicy)
        self.source_code.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        self.source_code.setAcceptRichText(False)
        self.source_code.setObjectName("source_code")
        self.vboxlayout2.addWidget(self.source_code)
        self.vboxlayout1.addWidget(self.groupBox_4)
        self.stacks.addWidget(self.page_2)
        self.verticalLayout.addWidget(self.stacks)
        self.horizontalLayout.addWidget(self.frame)
        self.verticalLayout_3.addWidget(self.central_widget)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout_4.addWidget(self.scrollArea)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_4.addWidget(self.buttonBox)
        self.label_2.setBuddy(self.profile_title)
        self.label_6.setBuddy(self.oldest_article)
        self.label_7.setBuddy(self.max_articles)
        self.label_4.setBuddy(self.feed_title)
        self.label_5.setBuddy(self.feed_url)

        self.retranslateUi(Dialog)
        self.stacks.setCurrentIndex(0)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):

        Dialog.setWindowTitle(_("Add custom news source"))
        self.groupBox.setTitle(_("Available user recipes"))
        self.add_profile_button.setText(_("Add/Update &recipe"))
        self.remove_profile_button.setText(_("&Remove recipe"))
        self.share_button.setText(_("&Share recipe"))
        self.show_recipe_files_button.setText(_("S&how recipe files"))
        self.builtin_recipe_button.setText(_("Customize &builtin recipe"))
        self.load_button.setText(_("&Load recipe from file"))
        self.opml_button.setToolTip(_("Import a collection of RSS feeds in OPML format\n"
"Many RSS readers can export their subscribed RSS feeds\n"
"in OPML format"))
        self.opml_button.setText(_("Import &OPML"))
        self.toggle_mode_button.setText(_("Switch to Advanced mode"))
        self.label.setText(_("<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'DejaVu Sans\'; font-size:10pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Create a basic news recipe, by adding RSS feeds to it. <br />For most feeds, you will have to use the \"Advanced mode\" to further customize the fetch process.</p></body></html>"))
        self.label_2.setText(_("Recipe &title:"))
        self.label_6.setText(_("&Oldest article:"))
        self.oldest_article.setToolTip(_("The oldest article to download"))
        self.oldest_article.setSuffix(_(" days"))
        self.label_7.setText(_("&Max. number of articles per feed:"))
        self.max_articles.setToolTip(_("Maximum number of articles to download per feed."))
        self.groupBox_2.setTitle(_("Feeds in recipe"))
        self.up_button.setText(_("..."))
        self.remove_feed_button.setToolTip(_("Remove feed from recipe"))
        self.remove_feed_button.setText(_("..."))
        self.down_button.setText(_("..."))
        self.groupBox_3.setTitle(_("Add feed to recipe"))
        self.label_4.setText(_("&Feed title:"))
        self.label_5.setText(_("Feed &URL:"))
        self.add_feed_button.setToolTip(_("Add feed to recipe"))
        self.add_feed_button.setText(_("&Add feed"))
        self.label_8.setText(_("For help with writing advanced news recipes, please visit <a href=\"http://manual.calibre-ebook.com/news.html\">User Recipes</a>"))
        self.groupBox_4.setTitle(_("Recipe source code (python)"))

from calibre.gui2.widgets import EnLineEdit, BasicList




# -*- coding: utf-8 -*-

__license__ = 'GPL 3'
__copyright__ = '2009, John Schember <john@nachtimwald.com>'
__docformat__ = 'restructuredtext en'


from calibre.gui2.convert.txt_output_ui import Ui_Form
from calibre.gui2.convert import Widget

class PluginWidget(Widget, Ui_Form):

    TITLE = _('TXT Output')
    HELP = _('Options specific to')+' TXT '+_('output')
    COMMIT_NAME = 'txt_output'
    ICON = I('mimetypes/txt.png')

    def __init__(self, parent, get_option, get_help, db=None, book_id=None):
        Widget.__init__(self, parent,
        ['newline', 'max_line_length', 'force_max_line_length',
        'inline_toc', 'txt_output_formatting', 'keep_links', 'keep_image_references',
        'keep_color', 'txt_output_encoding'])
        self.db, self.book_id = db, book_id
        for x in get_option('newline').option.choices:
            self.opt_newline.addItem(x)
        for x in get_option('txt_output_formatting').option.choices:
            self.opt_txt_output_formatting.addItem(x)
        self.initialize_options(get_option, get_help, db, book_id)

    def break_cycles(self):
        Widget.break_cycles(self)



from __future__ import with_statement
__license__   = 'GPL v3'
__copyright__ = '2008, Kovid Goyal kovid@kovidgoyal.net'
__docformat__ = 'restructuredtext en'

'''
Convert any ebook format to LIT.
'''

import sys, os, glob, logging

from calibre.ebooks.epub.from_any import any2epub, formats, USAGE
from calibre.ebooks.epub import config as common_config
from calibre.ptempfile import TemporaryDirectory
from calibre.ebooks.lit.writer import oeb2lit

def config(defaults=None):
    c = common_config(defaults=defaults, name='lit')
    return c

def option_parser(usage=USAGE):
    return config().option_parser(usage=usage%('LIT', formats()))

def any2lit(opts, path):
    ext = os.path.splitext(path)[1]
    if not ext:
        raise ValueError('Unknown file type: '+path)
    ext = ext.lower()[1:]
    
    if opts.output is None:
        opts.output = os.path.splitext(os.path.basename(path))[0]+'.lit'
    
    opts.output = os.path.abspath(opts.output)
    orig_output = opts.output
    
    with TemporaryDirectory('_any2lit') as tdir:
        oebdir = os.path.join(tdir, 'oeb')
        os.mkdir(oebdir)
        opts.output = os.path.join(tdir, 'dummy.epub')
        opts.profile = 'None'
        orig_bfs = opts.base_font_size2
        opts.base_font_size2 = 0
        any2epub(opts, path, create_epub=False, oeb_cover=True, extract_to=oebdir)
        opts.base_font_size2 = orig_bfs
        opf = glob.glob(os.path.join(oebdir, '*.opf'))[0]
        opts.output = orig_output
        logging.getLogger('html2epub').info(_('Creating LIT file from EPUB...'))
        oeb2lit(opts, opf)
    

def main(args=sys.argv):
    parser = option_parser()
    opts, args = parser.parse_args(args)
    if len(args) < 2:
        parser.print_help()
        print 'No input file specified.'
        return 1
    any2lit(opts, args[1])
    return 0

if __name__ == '__main__':
    sys.exit(main())



from calibre.web.feeds.news import BasicNewsRecipe

class AdvancedUserRecipe1278347258(BasicNewsRecipe):
    title      = u'Salt Lake City Tribune'
    __author__ = 'Charles Holbert'
    oldest_article = 7
    max_articles_per_feed = 100

    description            = '''Utah's independent news source since 1871'''
    publisher              = 'http://www.sltrib.com/'
    category               = 'news, Utah, SLC'
    language               = 'en'
    encoding               = 'utf-8'
    #delay                  = 1
    #simultaneous_downloads = 1
    remove_javascript      = True
    use_embedded_content   = False
    no_stylesheets         = True

    #masthead_url = 'http://www.sltrib.com/csp/cms/sites/sltrib/assets/images/logo_main.png'
    #cover_url = 'http://webmedia.newseum.org/newseum-multimedia/dfp/jpg9/lg/UT_SLT.jpg'

    keep_only_tags = [dict(name='div',attrs={'id':'imageBox'})
                      ,dict(name='div',attrs={'class':'headline'})
                      ,dict(name='div',attrs={'class':'byline'})
                      ,dict(name='p',attrs={'class':'TEXT_w_Indent'})]

    feeds = [(u'SL Tribune Today', u'http://www.sltrib.com/csp/cms/sites/sltrib/RSS/rss.csp?cat=All'),
           (u'Utah News', u'http://www.sltrib.com/csp/cms/sites/sltrib/RSS/rss.csp?cat=UtahNews'),
           (u'Business News', u'http://www.sltrib.com/csp/cms/sites/sltrib/RSS/rss.csp?cat=Money'),
           (u'Technology', u'http://www.sltrib.com/csp/cms/sites/sltrib/RSS/rss.csp?cat=Technology'),
           (u'Most Popular', u'http://www.sltrib.com/csp/cms/sites/sltrib/RSS/rsspopular.csp'),
           (u'Sports', u'http://www.sltrib.com/csp/cms/sites/sltrib/RSS/rss.csp?cat=Sports')]

    extra_css = '''
                .headline{font-family:Arial,Helvetica,sans-serif; font-size:xx-large; font-weight: bold; color:#0E5398;}
                .byline{font-family:Arial,Helvetica,sans-serif; color:#333333; font-size:xx-small;}
                .storytext{font-family:Arial,Helvetica,sans-serif; font-size:medium;}
                '''

    def print_version(self, url):
        seg = url.split('/')
        x = seg[5].split('-')
        baseURL = 'http://www.sltrib.com/csp/cms/sites/sltrib/pages/printerfriendly.csp?id='
        s = baseURL + x[0]
        return s

    def get_cover_url(self):
        cover_url = None
        href =  'http://www.newseum.org/todaysfrontpages/hr.asp?fpVname=UT_SLT&ref_pge=lst'
        soup = self.index_to_soup(href)
        div = soup.find('div',attrs={'class':'tfpLrgView_container'})
        if div:
            cover_url = div.img['src']
        return cover_url




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/dialogs/test_email.ui'
#
# Created: Fri Aug  8 18:32:27 2014
#      by: PyQt5 UI code generator 5.3.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(542, 418)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(I("config.png")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        Dialog.setWindowIcon(icon)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.from_ = QtWidgets.QLabel(Dialog)
        self.from_.setWordWrap(True)
        self.from_.setObjectName("from_")
        self.verticalLayout.addWidget(self.from_)
        self.to = QtWidgets.QLineEdit(Dialog)
        self.to.setObjectName("to")
        self.verticalLayout.addWidget(self.to)
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setText("")
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.test_button = QtWidgets.QPushButton(Dialog)
        self.test_button.setObjectName("test_button")
        self.verticalLayout.addWidget(self.test_button)
        self.log = QtWidgets.QPlainTextEdit(Dialog)
        self.log.setObjectName("log")
        self.verticalLayout.addWidget(self.log)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):

        Dialog.setWindowTitle(_("Test email settings"))
        self.from_.setText(_("Send test mail from %s to:"))
        self.test_button.setText(_("&Test"))


